"""
Conversion of :obj:`scmdata.ScmRun` into FaIR emissions :obj:`np.ndarray`.
"""
import datetime as dt
import os

import numpy as np
import pandas as pd
from scmdata import ScmRun


class HistoricalWorldEmms:
    """Historical world emissions class"""

    def __init__(self):
        self._loaded = False
        self._loaded_fair_history = False
        self._values = None
        self._values_fair_history = None

    @property
    def values(self):
        """Emissions values from historical joint with ssp245"""
        if not self._loaded:
            tmp = ScmRun(
                os.path.join(
                    os.path.dirname(__file__),
                    "rcmip-emissions-annual-means-v5-1-0-historical-ssp245.csv",
                ),
                lowercase_cols=True,
            )
            self._values = tmp.interpolate([dt.datetime(y, 1, 1) for y in tmp["year"]])

        self._loaded = True
        return self._values

    @property
    def values_fair_units(self):
        """Get values with FaIR's expected units, ignores species not used by FaIR"""
        if not self._loaded_fair_history:
            ssp_df_hist = self.values
            for variable in ssp_df_hist.get_unique_meta("variable"):
                in_unit = ssp_df_hist.filter(variable=variable).get_unique_meta(
                    "unit", no_duplicates=True
                )

                try:
                    _, fair_unit, context = _get_fair_col_unit_context(variable)
                except AssertionError:
                    # FaIR does not model the variable
                    if variable in [
                        "Emissions|F-Gases|HFC|HFC152a",
                        "Emissions|F-Gases|HFC|HFC236fa",
                        "Emissions|F-Gases|HFC|HFC365mfc",
                        "Emissions|F-Gases|NF3",
                        "Emissions|F-Gases|PFC|C3F8",
                        "Emissions|F-Gases|PFC|C4F10",
                        "Emissions|F-Gases|PFC|C5F12",
                        "Emissions|F-Gases|PFC|C7F16",
                        "Emissions|F-Gases|PFC|C8F18",
                        "Emissions|F-Gases|PFC|cC4F8",
                        "Emissions|F-Gases|SO2F2",
                        "Emissions|Montreal Gases|CH2Cl2",
                        "Emissions|Montreal Gases|CHCl3",
                    ]:
                        continue

                if in_unit != fair_unit:
                    ssp_df_hist = ssp_df_hist.convert_unit(
                        fair_unit, variable=variable, context=context
                    )

            self._values_fair_history = ssp_df_hist

        self._loaded_fair_history = True

        return self._values_fair_history


historical_world_emms_holder = HistoricalWorldEmms()


EMISSIONS_SPECIES_UNITS_CONTEXT = pd.DataFrame(
    (  # in fair 1.6, order is important
        ("|CO2|MAGICC Fossil and Industrial", "GtC / yr", None),
        ("|CO2|MAGICC AFOLU", "GtC / yr", None),
        ("|CH4", "MtCH4 / yr", None),
        ("|N2O", "MtN2ON / yr", None),
        ("|Sulfur", "MtS / yr", None),
        ("|CO", "MtCO / yr", None),
        ("|VOC", "MtNMVOC / yr", None),
        ("|NOx", "MtN / yr", "NOx_conversions"),
        ("|BC", "MtBC / yr", None),
        ("|OC", "MtOC / yr", None),
        ("|NH3", "MtNH3 / yr", None),
        ("|CF4", "ktCF4 / yr", None),
        ("|C2F6", "ktC2F6 / yr", None),
        ("|C6F14", "ktC6F14 / yr", None),
        ("|HFC23", "ktHFC23 / yr", None),
        ("|HFC32", "ktHFC32 / yr", None),
        ("|HFC4310mee", "ktHFC4310mee / yr", None),
        ("|HFC125", "ktHFC125 / yr", None),
        ("|HFC134a", "ktHFC134a / yr", None),
        ("|HFC143a", "ktHFC143a / yr", None),
        ("|HFC227ea", "ktHFC227ea / yr", None),
        ("|HFC245fa", "ktHFC245fa / yr", None),
        ("|SF6", "ktSF6 / yr", None),
        ("|CFC11", "ktCFC11 / yr", None),
        ("|CFC12", "ktCFC12 / yr", None),
        ("|CFC113", "ktCFC113 / yr", None),
        ("|CFC114", "ktCFC114 / yr", None),
        ("|CFC115", "ktCFC115 / yr", None),
        ("|CCl4", "ktCCl4 / yr", None),
        ("|CH3CCl3", "ktCH3CCl3 / yr", None),
        ("|HCFC22", "ktHCFC22 / yr", None),
        ("|HCFC141b", "ktHCFC141b / yr", None),
        ("|HCFC142b", "ktHCFC142b / yr", None),
        ("|Halon1211", "ktHalon1211 / yr", None),
        ("|Halon1202", "ktHalon1202 / yr", None),
        ("|Halon1301", "ktHalon1301 / yr", None),
        ("|Halon2402", "ktHalon2402 / yr", None),
        ("|CH3Br", "ktCH3Br / yr", None),
        ("|CH3Cl", "ktCH3Cl / yr", None),
    ),
    columns=["species", "in_unit", "context"],
)


def _get_fair_col_unit_context(variable):
    row = EMISSIONS_SPECIES_UNITS_CONTEXT["species"].apply(
        lambda x: variable.endswith(x)  # pylint: disable=W0108
    )
    in_unit = EMISSIONS_SPECIES_UNITS_CONTEXT[row]["in_unit"]
    if in_unit.shape[0] != 1:
        raise AssertionError(in_unit)

    fair_col = int(row[row].index.values) + 1  # first col is time
    in_unit = in_unit.iloc[0]
    context = EMISSIONS_SPECIES_UNITS_CONTEXT[row]["context"].iloc[0]

    return fair_col, in_unit, context


def scmdf_to_emissions(
    scmrun, startyear=1750, endyear=2100, scen_startyear=2015
):  # pylint: disable=R0914
    """
    Convert an :obj:`scmdata.ScmRun` into a FaIR emissions :obj:`np.ndarray`

    Interpolates linearly if required and fills in montreal gases based on ssp245.
    between non-consecutive years in the SCEN file. Fills in Montreal gases
    from SSP2-4.5.

    Parameters
    ----------
    scmrun : :obj:`ScmRun`
        :obj:`ScmRun` instance to convert to FaIR emissions array.
    startyear : int
        First year of output array to produce.
    endyear : int
        Last year of output array to produce.
    scen_startyear : int
        First year of the future scenario corresponding to :obj:`ScmRun`.

    Returns
    -------
    np.ndarray
        FaIR emissions array of size (nt, 40).

    Raises
    ------
    AssertionError
        If there is more than one model-scenario pair in the provided :obj:`ScmRun`.
    """
    n_cols = 40
    nt = endyear - startyear + 1

    data_out = np.ones((nt, n_cols)) * np.nan
    data_out[:, 0] = np.arange(startyear, endyear + 1)

    if scmrun.meta[["model", "scenario"]].drop_duplicates().shape[0] != 1:
        raise AssertionError("Should only have one model-scenario pair")

    scmrun = scmrun.interpolate(
        [dt.datetime(y, 1, 1) for y in range(scen_startyear, endyear + 1)]
    )

    years = scmrun["year"].values
    first_scenyear = years[0]
    first_scen_row = int(first_scenyear - startyear)

    # if correct units and interpolation were guaranteed we could do this for scenario too which is quicker
    hist_df = historical_world_emms_holder.values_fair_units.filter(
        year=range(startyear, scen_startyear)
    ).timeseries()

    future_df = historical_world_emms_holder.values_fair_units.filter(
        year=range(scen_startyear, endyear + 1)
    ).timeseries()

    for species in EMISSIONS_SPECIES_UNITS_CONTEXT["species"]:
        fair_col, _, _ = _get_fair_col_unit_context(species)

        hist_df_row = hist_df.index.get_level_values("variable").str.endswith(species)

        data_out[:first_scen_row, fair_col] = hist_df[hist_df_row].values.squeeze()

        future_df_row = future_df.index.get_level_values("variable").str.endswith(
            species
        )
        data_out[first_scen_row:, fair_col] = future_df[future_df_row].values.squeeze()

    for var_df in scmrun.groupby("variable"):
        variable = var_df.get_unique_meta("variable", no_duplicates=True)

        # Skip aggregate emissions and emissons not handled by FaIR
        if (
            variable.split("Emissions")[1]
            not in EMISSIONS_SPECIES_UNITS_CONTEXT.species.values
        ):
            continue

        in_unit = var_df.get_unique_meta("unit", no_duplicates=True)

        fair_col, fair_unit, context = _get_fair_col_unit_context(variable)

        if in_unit != fair_unit:
            var_df_fair_unit = var_df.convert_unit(fair_unit, context=context)
        else:
            var_df_fair_unit = var_df

        data_out[first_scen_row:, fair_col] = var_df_fair_unit.values.squeeze()

    return data_out



"""
Module for running CICEROSCM in parallel
"""
import logging
import os
from concurrent.futures import ProcessPoolExecutor

import scmdata

from ...settings import config
from ..utils._parallel_process import _parallel_process
from .ciceroscm_wrapper import CiceroSCMWrapper

LOGGER = logging.getLogger(__name__)


FRONT_SERIAL = 0
"""int: Number of serial runs to do before starting parallel runs"""

FRONT_PARALLEL = 0
"""int: Number of front parallel runs to do before starting full parallel runs"""


def _execute_run(cfgs, output_variables, scenariodata):
    cscm = CiceroSCMWrapper(scenariodata)
    try:
        out = cscm.run_over_cfgs(cfgs, output_variables)
    finally:
        cscm.cleanup_tempdirs()

    return out


def run_ciceroscm_parallel(scenarios, cfgs, output_vars):
    """
    Run CICEROSCM in parallel

    Parameters
    ----------
    scenarios : IamDataFrame
        Scenariodata with which to run

    cfgs : list[dict]
        List of configurations with which to run CICEROSCM

    output_vars : list[str]
        Variables to output (may require some fiddling with ``out_x``
        variables in ``cfgs`` to get this right)

    Returns
    -------
    :obj:`ScmRun`
        :obj:`ScmRun` instance with all results.
    """
    LOGGER.info("Entered _parallel_ciceroscm")
    runs = [
        {"cfgs": cfgs, "output_variables": output_vars, "scenariodata": smdf}
        for (scen, model), smdf in scenarios.timeseries().groupby(["scenario", "model"])
    ]

    max_workers = int(config.get("CICEROSCM_WORKER_NUMBER", os.cpu_count()))
    LOGGER.info("Running in parallel with up to %d workers", max_workers)

    with ProcessPoolExecutor(max_workers=max_workers) as pool:
        result = _parallel_process(
            func=_execute_run,
            configuration=runs,
            pool=pool,
            config_are_kwargs=True,
            # no front runs as these defeat the purpose with CICERO-SCM (because
            # it is only parallel on scenarios, not configs)
            front_serial=FRONT_SERIAL,
            front_parallel=FRONT_PARALLEL,
        )

    LOGGER.info("Appending CICERO-SCM results into a single ScmRun")
    result = scmdata.run_append([r for r in result if r is not None])

    return result



"""
FAIR adapter
"""
import functools
import os

import numpy as np
import pandas as pd
from scmdata import ScmRun

from ...progress import progress
from ..base import _Adapter
from ._compat import fair
from ._run_fair import run_fair
from ._scmdf_to_emissions import scmdf_to_emissions


@functools.lru_cache()
def _get_natural_emissions_and_forcing(startyear, num_timesteps):
    # TODO: somebody who knows what they are doing to use scmdata
    natural_df = pd.read_csv(
        os.path.join(os.path.dirname(__file__), "natural-emissions-and-forcing.csv",),
    )
    ndf_values = natural_df.values

    n_index_columns = 7
    start_index = startyear - 1750 + n_index_columns
    ch4_n2o = ndf_values[0:2, start_index : start_index + num_timesteps].T
    solar_forcing = ndf_values[2, start_index : start_index + num_timesteps].T
    volcanic_forcing = ndf_values[3, start_index : start_index + num_timesteps].T

    return {
        "ch4_n2o": ch4_n2o,
        "solar_forcing": solar_forcing,
        "volcanic_forcing": volcanic_forcing,
    }


class FAIR(_Adapter):
    """
    Adapter for running FAIR
    """

    def _init_model(self, *args, **kwargs):
        if fair is None:
            raise ImportError("fair is not installed. Run 'pip install fair'")

    def _run(self, scenarios, cfgs, output_variables, output_config):
        if output_config is not None:
            raise NotImplementedError("`output_config` not implemented for FaIR")

        fair_df = ScmRun(scenarios.timeseries())
        full_cfgs = self._make_full_cfgs(fair_df, cfgs)

        res = run_fair(full_cfgs, output_variables)
        res["climate_model"] = f"FaIRv{self.get_version()}"

        return res

    def _make_full_cfgs(self, scenarios, cfgs):  # pylint: disable=R0201,R0914
        full_cfgs = []
        run_id_block = 0
        startyear = _check_startyear(cfgs)

        for (scenario, model), smdf in progress(
            scenarios.timeseries().groupby(["scenario", "model"]),
            desc="Creating FaIR emissions inputs",
        ):
            smdf_in = ScmRun(smdf)
            scen_startyear = smdf_in.time_points.years()[0]
            endyear = smdf_in.time_points.years()[-1]

            if startyear < 1750:
                raise ValueError(
                    f"startyear must be 1750 or later ({startyear} specified)"
                )
            if endyear > 2500:
                raise ValueError(
                    f"endyear must be 2500 or earlier ({endyear} implied by scenario data)"
                )
            emissions = scmdf_to_emissions(
                smdf_in,
                startyear=startyear,
                scen_startyear=scen_startyear,
                endyear=endyear,
            )
            num_timesteps = emissions.shape[0]

            natural_components = _get_natural_emissions_and_forcing(
                startyear, num_timesteps
            )
            ch4_n2o = natural_components["ch4_n2o"]
            solar_forcing = natural_components["solar_forcing"]
            volcanic_forcing = natural_components["volcanic_forcing"]

            scenario_cfg = [
                {
                    "scenario": scenario,
                    "model": model,
                    "run_id": run_id_block + i,
                    "emissions": emissions,
                    "natural": ch4_n2o,
                    "F_volcanic": volcanic_forcing,
                    "F_solar": solar_forcing,
                    "efficacy": np.ones(45),
                    "diagnostics": "AR6",
                    "gir_carbon_cycle": True,
                    "temperature_function": "Geoffroy",
                    "aerosol_forcing": "aerocom+ghan2",
                    "fixPre1850RCP": False,
                    "b_tro3": np.array(
                        [1.77871043e-04, 5.80173377e-05, 1.94458719e-04, 2.09151270e-03]
                    ),
                    "tropO3_forcing": "cmip6",
                    "aCO2land": 0.0006394631886297174,
                    "b_aero": np.array([-0.00503, 0.0, 0.0, 0.0, 0.0385, -0.0104, 0.0]),
                    "ghan_params": np.array([1.232, 73.9, 63.0]),
                    "gmst_factor": 1 / 1.04,
                    "ohu_factor": 0.92,
                    "startyear": startyear,
                    **cfg,
                }
                for i, cfg in enumerate(cfgs)
            ]
            run_id_block += len(scenario_cfg)

            full_cfgs += scenario_cfg

        return full_cfgs

    @staticmethod
    def get_version():
        """
        Get the FAIR version being used by this adapter
        Returns
        -------
        str
            The FAIR version id
        """
        return fair.__version__


def _check_startyear(cfgs):
    """
    Check to see that at most one startyear is defined in the config
    Returns
    -------
    int
        startyear

    Raises
    ------
    ValueError
        if more that one startyear is defined
    """
    first_startyear = cfgs[0].pop("startyear", 1750)
    if len(cfgs) > 1:
        for cfg in cfgs[1:]:
            this_startyear = cfg.pop("startyear", 1750)
            if this_startyear != first_startyear:
                raise ValueError("Can only handle one startyear per scenario ensemble")

    return first_startyear



import os.path

import numpy as np
import numpy.testing as npt
import pytest
from base import _AdapterTester
from scmdata import ScmRun

from openscm_runner import run
from openscm_runner.adapters import FAIR
from openscm_runner.utils import calculate_quantiles


class TestFairAdapter(_AdapterTester):
    @pytest.mark.parametrize("nworkers", (1, 4))
    def test_run(
        self,
        test_scenarios,
        monkeypatch,
        nworkers,
        test_data_dir,
        update_expected_values,
    ):
        expected_output_file = os.path.join(
            test_data_dir,
            "expected-integration-output",
            "expected_fair1X_test_run_output.json",
        )

        monkeypatch.setenv("FAIR_WORKER_NUMBER", "{}".format(nworkers))
        res = run(
            climate_models_cfgs={
                "FaIR": [
                    {},
                    {"q": np.array([0.3, 0.45]), "r0": 30.0, "lambda_global": 0.9},
                    {"q": np.array([0.35, 0.4]), "r0": 25.0, "lambda_global": 1.1},
                ],
            },
            scenarios=test_scenarios.filter(scenario=["ssp126", "ssp245", "ssp370"]),
            output_variables=(
                "Surface Air Temperature Change",
                "Atmospheric Concentrations|CO2",
                "Heat Content",
                "Heat Uptake",
                "Effective Radiative Forcing",
                "Effective Radiative Forcing|Aerosols",
                "Effective Radiative Forcing|CO2",
                "CO2 Air to Land Flux",  # should be ignored
            ),
            out_config=None,
        )

        assert isinstance(res, ScmRun)
        assert res["run_id"].min() == 0
        assert res["run_id"].max() == 8
        assert res.get_unique_meta(
            "climate_model", no_duplicates=True
        ) == "FaIRv{}".format(FAIR.get_version())

        assert set(res.get_unique_meta("variable")) == set(
            [
                "Surface Air Temperature Change",
                "Atmospheric Concentrations|CO2",
                "Heat Content",
                "Heat Uptake",
                "Effective Radiative Forcing",
                "Effective Radiative Forcing|Aerosols",
                "Effective Radiative Forcing|CO2",
            ]
        )

        # check we can also calcluate quantiles
        assert "run_id" in res.meta
        quantiles = calculate_quantiles(res, [0, 0.05, 0.17, 0.5, 0.83, 0.95, 1])
        assert "run_id" not in quantiles.meta

        # TODO CHECK: heat content is not zero in the first year in FaIR?
        self._check_heat_content_heat_uptake_consistency(res)

        self._check_output(res, expected_output_file, update_expected_values)

    def test_variable_naming(self, test_scenarios):
        missing_from_fair = (
            "Effective Radiative Forcing|Aerosols|Direct Effect|BC|MAGICC AFOLU",
            "Effective Radiative Forcing|Aerosols|Direct Effect|BC|MAGICC Fossil and Industrial",
            "Effective Radiative Forcing|Aerosols|Direct Effect|OC|MAGICC AFOLU",
            "Effective Radiative Forcing|Aerosols|Direct Effect|OC|MAGICC Fossil and Industrial",
            "Effective Radiative Forcing|Aerosols|Direct Effect|SOx|MAGICC AFOLU",
            "Effective Radiative Forcing|Aerosols|Direct Effect|SOx|MAGICC Fossil and Industrial",
            "Net Atmosphere to Ocean Flux|CO2",
            "Net Atmosphere to Land Flux|CO2",
        )
        common_variables = [
            c for c in self._common_variables if c not in missing_from_fair
        ]
        res = run(
            climate_models_cfgs={"FaIR": ({"startyear": 1750},)},
            scenarios=test_scenarios.filter(scenario="ssp126"),
            output_variables=common_variables,
        )

        missing_vars = set(common_variables) - set(res["variable"])
        if missing_vars:
            raise AssertionError(missing_vars)


def test_fair_ocean_factors(test_scenarios):
    res_default_factors = run(
        climate_models_cfgs={"FaIR": [{}]},
        scenarios=test_scenarios.filter(scenario=["ssp585"]),
        output_variables=(
            "Surface Air Ocean Blended Temperature Change",
            "Heat Uptake|Ocean",
            "Heat Content|Ocean",
        ),
    )

    res_custom_factors = run(
        climate_models_cfgs={
            "FaIR": [
                {
                    "gmst_factor": np.linspace(0.90, 1.00, 351),  # test with array
                    "ohu_factor": 0.93,
                }
            ]
        },
        scenarios=test_scenarios.filter(scenario=["ssp585"]),
        output_variables=(
            "Surface Air Ocean Blended Temperature Change",
            "Heat Uptake|Ocean",
            "Heat Content|Ocean",
        ),
    )

    assert (
        res_default_factors.filter(
            variable="Surface Air Ocean Blended Temperature Change",
            region="World",
            year=2100,
            scenario="ssp585",
        ).values
        != res_custom_factors.filter(
            variable="Surface Air Ocean Blended Temperature Change",
            region="World",
            year=2100,
            scenario="ssp585",
        ).values
    )


def test_startyear(test_scenarios, test_scenarios_2600):
    # we can't run different start years in the same ensemble as output files will differ in shape.
    # There is a separate test to ensure this does raise an error.
    res_1850 = run(
        climate_models_cfgs={"FaIR": [{"startyear": 1850}]},
        scenarios=test_scenarios.filter(scenario=["ssp245"]),
        output_variables=("Surface Air Temperature Change",),
        out_config=None,
    )

    res_1750 = run(
        climate_models_cfgs={"FaIR": [{"startyear": 1750}]},
        scenarios=test_scenarios.filter(scenario=["ssp245"]),
        output_variables=("Surface Air Temperature Change",),
        out_config=None,
    )

    res_default = run(
        climate_models_cfgs={"FaIR": [{}]},
        scenarios=test_scenarios.filter(scenario=["ssp245"]),
        output_variables=("Surface Air Temperature Change",),
        out_config=None,
    )

    gsat2100_start1850 = res_1850.filter(
        variable="Surface Air Temperature Change", region="World", year=2100,
    ).values

    gsat2100_start1750 = res_1750.filter(
        variable="Surface Air Temperature Change", region="World", year=2100,
    ).values

    gsat2100_startdefault = res_default.filter(
        variable="Surface Air Temperature Change", region="World", year=2100,
    ).values

    assert gsat2100_start1850 != gsat2100_start1750
    assert gsat2100_start1750 == gsat2100_startdefault

    with pytest.raises(ValueError):
        run(
            climate_models_cfgs={"FaIR": [{"startyear": 1650}]},
            scenarios=test_scenarios.filter(scenario=["ssp245"]),
            output_variables=("Surface Air Temperature Change",),
            out_config=None,
        )

    with pytest.raises(ValueError):
        run(
            climate_models_cfgs={"FaIR": [{}]},
            scenarios=test_scenarios_2600.filter(scenario=["ssp245"]),
            output_variables=("Surface Air Temperature Change",),
            out_config=None,
        )

    with pytest.raises(ValueError):
        run(
            climate_models_cfgs={"FaIR": [{"startyear": 1750}, {"startyear": 1850}]},
            scenarios=test_scenarios.filter(scenario=["ssp245"]),
            output_variables=("Surface Air Temperature Change",),
            out_config=None,
        )


def test_forcing_categories(test_scenarios):
    forcing_categories = [
        "Effective Radiative Forcing|CO2",
        "Effective Radiative Forcing|CH4",
        "Effective Radiative Forcing|N2O",
        "Effective Radiative Forcing|CF4",
        "Effective Radiative Forcing|C2F6",
        "Effective Radiative Forcing|C6F14",
        "Effective Radiative Forcing|HFC23",
        "Effective Radiative Forcing|HFC32",
        "Effective Radiative Forcing|HFC125",
        "Effective Radiative Forcing|HFC134a",
        "Effective Radiative Forcing|HFC143a",
        "Effective Radiative Forcing|HFC227ea",
        "Effective Radiative Forcing|HFC245fa",
        "Effective Radiative Forcing|HFC4310mee",
        "Effective Radiative Forcing|SF6",
        "Effective Radiative Forcing|CFC11",
        "Effective Radiative Forcing|CFC12",
        "Effective Radiative Forcing|CFC113",
        "Effective Radiative Forcing|CFC114",
        "Effective Radiative Forcing|CFC115",
        "Effective Radiative Forcing|CCl4",
        "Effective Radiative Forcing|CH3CCl3",
        "Effective Radiative Forcing|HCFC22",
        "Effective Radiative Forcing|HCFC141b",
        "Effective Radiative Forcing|HCFC142b",
        "Effective Radiative Forcing|Halon1211",
        "Effective Radiative Forcing|Halon1202",
        "Effective Radiative Forcing|Halon1301",
        "Effective Radiative Forcing|Halon2402",
        "Effective Radiative Forcing|CH3Br",
        "Effective Radiative Forcing|CH3Cl",
        "Effective Radiative Forcing|Tropospheric Ozone",
        "Effective Radiative Forcing|Stratospheric Ozone",
        "Effective Radiative Forcing|CH4 Oxidation Stratospheric H2O",
        "Effective Radiative Forcing|Contrails",
        "Effective Radiative Forcing|Aerosols|Direct Effect|SOx",
        "Effective Radiative Forcing|Aerosols|Direct Effect|Secondary Organic Aerosol",
        "Effective Radiative Forcing|Aerosols|Direct Effect|Nitrate",
        "Effective Radiative Forcing|Aerosols|Direct Effect|BC",
        "Effective Radiative Forcing|Aerosols|Direct Effect|OC",
        "Effective Radiative Forcing|Aerosols|Indirect Effect",
        "Effective Radiative Forcing|Black Carbon on Snow",
        "Effective Radiative Forcing|Land-use Change",
        "Effective Radiative Forcing|Volcanic",
        "Effective Radiative Forcing|Solar",
        "Effective Radiative Forcing",
        "Effective Radiative Forcing|Anthropogenic",
        "Effective Radiative Forcing|Greenhouse Gases",
        "Effective Radiative Forcing|Kyoto Gases",
        "Effective Radiative Forcing|CO2, CH4 and N2O",
        "Effective Radiative Forcing|F-Gases",
        "Effective Radiative Forcing|Montreal Protocol Halogen Gases",
        "Effective Radiative Forcing|Aerosols|Direct Effect",
        "Effective Radiative Forcing|Aerosols",
        "Effective Radiative Forcing|Ozone",
        "Effective Radiative Forcing",
    ]

    res = run(
        climate_models_cfgs={"FaIR": [{}]},
        scenarios=test_scenarios.filter(scenario=["ssp245"]),
        output_variables=tuple(forcing_categories),
        out_config=None,
    )

    # storing results in a dict makes this a bit more compact
    forcing = {}
    for variable in forcing_categories:
        forcing[variable] = res.filter(variable=variable, region="World").values

    npt.assert_allclose(
        forcing["Effective Radiative Forcing|CO2"]
        + forcing["Effective Radiative Forcing|CH4"]
        + forcing["Effective Radiative Forcing|N2O"]
        + forcing["Effective Radiative Forcing|CF4"]
        + forcing["Effective Radiative Forcing|C2F6"]
        + forcing["Effective Radiative Forcing|C6F14"]
        + forcing["Effective Radiative Forcing|HFC23"]
        + forcing["Effective Radiative Forcing|HFC32"]
        + forcing["Effective Radiative Forcing|HFC125"]
        + forcing["Effective Radiative Forcing|HFC134a"]
        + forcing["Effective Radiative Forcing|HFC143a"]
        + forcing["Effective Radiative Forcing|HFC227ea"]
        + forcing["Effective Radiative Forcing|HFC245fa"]
        + forcing["Effective Radiative Forcing|HFC4310mee"]
        + forcing["Effective Radiative Forcing|SF6"]
        + forcing["Effective Radiative Forcing|CFC11"]
        + forcing["Effective Radiative Forcing|CFC12"]
        + forcing["Effective Radiative Forcing|CFC113"]
        + forcing["Effective Radiative Forcing|CFC114"]
        + forcing["Effective Radiative Forcing|CFC115"]
        + forcing["Effective Radiative Forcing|CCl4"]
        + forcing["Effective Radiative Forcing|CH3CCl3"]
        + forcing["Effective Radiative Forcing|HCFC22"]
        + forcing["Effective Radiative Forcing|HCFC141b"]
        + forcing["Effective Radiative Forcing|HCFC142b"]
        + forcing["Effective Radiative Forcing|Halon1211"]
        + forcing["Effective Radiative Forcing|Halon1202"]
        + forcing["Effective Radiative Forcing|Halon1301"]
        + forcing["Effective Radiative Forcing|Halon2402"]
        + forcing["Effective Radiative Forcing|CH3Br"]
        + forcing["Effective Radiative Forcing|CH3Cl"],
        forcing[
            "Effective Radiative Forcing|Greenhouse Gases"
        ],  # should this be "well mixed" greenhouse gases?
    )

    npt.assert_allclose(
        forcing["Effective Radiative Forcing|CO2"]
        + forcing["Effective Radiative Forcing|CH4"]
        + forcing["Effective Radiative Forcing|N2O"]
        + forcing["Effective Radiative Forcing|CF4"]
        + forcing["Effective Radiative Forcing|C2F6"]
        + forcing["Effective Radiative Forcing|C6F14"]
        + forcing["Effective Radiative Forcing|HFC23"]
        + forcing["Effective Radiative Forcing|HFC32"]
        + forcing["Effective Radiative Forcing|HFC125"]
        + forcing["Effective Radiative Forcing|HFC134a"]
        + forcing["Effective Radiative Forcing|HFC143a"]
        + forcing["Effective Radiative Forcing|HFC227ea"]
        + forcing["Effective Radiative Forcing|HFC245fa"]
        + forcing["Effective Radiative Forcing|HFC4310mee"]
        + forcing["Effective Radiative Forcing|SF6"],
        forcing["Effective Radiative Forcing|Kyoto Gases"],
    )

    npt.assert_allclose(
        forcing["Effective Radiative Forcing|CO2"]
        + forcing["Effective Radiative Forcing|CH4"]
        + forcing["Effective Radiative Forcing|N2O"],
        forcing["Effective Radiative Forcing|CO2, CH4 and N2O"],
    )

    npt.assert_allclose(
        forcing["Effective Radiative Forcing|CF4"]
        + forcing["Effective Radiative Forcing|C2F6"]
        + forcing["Effective Radiative Forcing|C6F14"]
        + forcing["Effective Radiative Forcing|HFC23"]
        + forcing["Effective Radiative Forcing|HFC32"]
        + forcing["Effective Radiative Forcing|HFC125"]
        + forcing["Effective Radiative Forcing|HFC134a"]
        + forcing["Effective Radiative Forcing|HFC143a"]
        + forcing["Effective Radiative Forcing|HFC227ea"]
        + forcing["Effective Radiative Forcing|HFC245fa"]
        + forcing["Effective Radiative Forcing|HFC4310mee"]
        + forcing["Effective Radiative Forcing|SF6"],
        forcing["Effective Radiative Forcing|F-Gases"],
    )

    npt.assert_allclose(
        forcing["Effective Radiative Forcing|CFC11"]
        + forcing["Effective Radiative Forcing|CFC12"]
        + forcing["Effective Radiative Forcing|CFC113"]
        + forcing["Effective Radiative Forcing|CFC114"]
        + forcing["Effective Radiative Forcing|CFC115"]
        + forcing["Effective Radiative Forcing|CCl4"]
        + forcing["Effective Radiative Forcing|CH3CCl3"]
        + forcing["Effective Radiative Forcing|HCFC22"]
        + forcing["Effective Radiative Forcing|HCFC141b"]
        + forcing["Effective Radiative Forcing|HCFC142b"]
        + forcing["Effective Radiative Forcing|Halon1211"]
        + forcing["Effective Radiative Forcing|Halon1202"]
        + forcing["Effective Radiative Forcing|Halon1301"]
        + forcing["Effective Radiative Forcing|Halon2402"]
        + forcing["Effective Radiative Forcing|CH3Br"]
        + forcing["Effective Radiative Forcing|CH3Cl"],
        forcing["Effective Radiative Forcing|Montreal Protocol Halogen Gases"],
    )

    npt.assert_allclose(
        forcing["Effective Radiative Forcing|Tropospheric Ozone"]
        + forcing["Effective Radiative Forcing|Stratospheric Ozone"],
        forcing["Effective Radiative Forcing|Ozone"],
    )

    npt.assert_allclose(
        forcing["Effective Radiative Forcing|Aerosols|Direct Effect|SOx"]
        + forcing[
            "Effective Radiative Forcing|Aerosols|Direct Effect|Secondary Organic Aerosol"
        ]
        + forcing["Effective Radiative Forcing|Aerosols|Direct Effect|Nitrate"]
        + forcing["Effective Radiative Forcing|Aerosols|Direct Effect|BC"]
        + forcing["Effective Radiative Forcing|Aerosols|Direct Effect|OC"],
        forcing["Effective Radiative Forcing|Aerosols|Direct Effect"],
    )

    # If up to here is fine, then we only need to check previouly defined aggregates against "super-aggregates"
    npt.assert_allclose(
        forcing["Effective Radiative Forcing|Aerosols|Direct Effect"]
        + forcing["Effective Radiative Forcing|Aerosols|Indirect Effect"],
        forcing["Effective Radiative Forcing|Aerosols"],
    )

    npt.assert_allclose(
        forcing["Effective Radiative Forcing|Greenhouse Gases"]
        + forcing["Effective Radiative Forcing|Ozone"]
        + forcing["Effective Radiative Forcing|CH4 Oxidation Stratospheric H2O"]
        + forcing["Effective Radiative Forcing|Contrails"]
        + forcing["Effective Radiative Forcing|Aerosols"]
        + forcing["Effective Radiative Forcing|Black Carbon on Snow"]
        + forcing["Effective Radiative Forcing|Land-use Change"],
        forcing["Effective Radiative Forcing|Anthropogenic"],
    )

    npt.assert_allclose(
        forcing["Effective Radiative Forcing|Anthropogenic"]
        + forcing["Effective Radiative Forcing|Volcanic"]
        + forcing["Effective Radiative Forcing|Solar"],
        forcing["Effective Radiative Forcing"],
    )



from unittest.mock import patch

import pytest

from openscm_runner import run


@patch("openscm_runner.adapters.fair_adapter.fair_adapter.fair", None)
def test_no_fair():
    with pytest.raises(
        ImportError, match="fair is not installed. Run 'pip install fair'"
    ):
        run(
            climate_models_cfgs={"fair": ["config list"]}, scenarios="not used",
        )


@patch("openscm_runner.adapters.magicc7.magicc7.pymagicc", None)
def test_no_pymagicc():
    with pytest.raises(
        ImportError,
        match="pymagicc is not installed. Run 'conda install pymagicc' or 'pip install pymagicc'",
    ):
        run(
            climate_models_cfgs={"MAGICC7": ["config list"]}, scenarios="not used",
        )



import os.path

import pyam
import pytest
from scmdata import ScmRun

from openscm_runner.adapters import CICEROSCM, MAGICC7

TEST_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "test-data")


@pytest.fixture(scope="session")
def test_data_dir():
    return TEST_DATA_DIR


@pytest.fixture(scope="session")
def test_scenarios(test_data_dir):
    scenarios = pyam.IamDataFrame(
        os.path.join(test_data_dir, "rcmip_scen_ssp_world_emissions.csv")
    )

    return scenarios


@pytest.fixture(scope="session")
def test_scenarios_2600(test_data_dir):
    scenarios = pyam.IamDataFrame(
        os.path.join(test_data_dir, "rcmip_scen_ssp_world_emissions_2600.csv")
    )

    return scenarios


@pytest.fixture(scope="session")
def test_scenario_ssp370_world(test_data_dir):
    scenario = ScmRun(
        os.path.join(
            test_data_dir, "rcmip-emissions-annual-means-v5-1-0-ssp370-world.csv"
        ),
        lowercase_cols=True,
    )

    return scenario


REQUIRED_MAGICC_VERSION = "v7.5.1"

try:
    MAGICC_VERSION = MAGICC7.get_version()
    if MAGICC_VERSION != REQUIRED_MAGICC_VERSION:
        CORRECT_MAGICC_IS_AVAILABLE = False
    else:
        CORRECT_MAGICC_IS_AVAILABLE = True

except KeyError:
    MAGICC_VERSION = None
    CORRECT_MAGICC_IS_AVAILABLE = False


CICEROSCM_IS_AVAILABLE = False
try:
    CICEROSCM.get_version()
    CICEROSCM_IS_AVAILABLE = True
except OSError:
    CICERO_OS_ERROR = True
except KeyError:
    CICERO_OS_ERROR = False


def pytest_runtest_setup(item):
    for mark in item.iter_markers():
        if mark.name == "magicc" and not CORRECT_MAGICC_IS_AVAILABLE:
            if MAGICC_VERSION is None:
                pytest.skip("MAGICC7 not available")
            else:
                pytest.skip(
                    "Wrong MAGICC version for tests ({}), we require {}".format(
                        MAGICC_VERSION, REQUIRED_MAGICC_VERSION
                    )
                )

        if mark.name == "ciceroscm" and not CICEROSCM_IS_AVAILABLE:
            if CICERO_OS_ERROR:
                pytest.skip("CICERO-SCM not available because of operating system")
            else:
                pytest.skip("CICERO-SCM not available")


def pytest_addoption(parser):
    parser.addoption(
        "--update-expected-values",
        action="store_true",
        default=False,
        help="Overwrite expected values",
    )


@pytest.fixture
def update_expected_values(request):
    return request.config.getoption("--update-expected-values")



"""
Module that makes CICERO-SCM parameter files
"""
import os
from pathlib import Path


def splitall(path):
    """
    Split path into each part, top catalogue on top, filename (if included) last
    """
    out = Path(path).parts
    return out


class PARAMETERFILEWRITER:  # pylint: disable=too-few-public-methods
    """
    Class to write parameterfiles
    """

    # pylint: disable=too-many-instance-attributes
    # Eight is reasonable in this case.

    def __init__(self, udir):
        self.udir = udir
        self._pamset_defaults = {
            "model_end": 2500,
            "scenario_start": 2015,
            "scenario_end": 2500,
            "lambda": "0.540",
            "akapa": "0.341",
            "cpi": "0.556",
            "W": "1.897",
            "rlamdo": "16.618",
            "beto": "3.225",
            "mixed": "107.277",
            "dirso2_forc": "-0.457",
            "indso2_forc": "-0.514",
            "bc_forc": "0.200",
            "oc_forc": "-0.103",
        }

        with open(
            os.path.join(self.udir, "pam_RCMIP_test_klimsensdefault.scm"),
            "r",
            encoding="ascii",
        ) as origfile:
            self.origfiledata = origfile.read()

    def write_parameterfile(self, pamset, filedir):
        """
        Make parameter file for single run
        """
        scen = splitall(filedir)[-1]
        filedir_to_pamfile = os.path.join(".", scen)
        filedata = self.origfiledata
        filedata = filedata.replace(
            "output_rbs/test_rcmip",
            os.path.join(filedir_to_pamfile, "outputfiles", "temp"),
        )
        filedata = filedata.replace("../input_RCP/", "")
        filedata = filedata.replace("input/ssp434_conc_", "ssp245_conc_")

        filedata = filedata.replace(
            "input/ssp434_em_RCMIP.txt",
            f"{filedir_to_pamfile}/inputfiles/{scen}_em.txt",
        )
        for k, value in self._pamset_defaults.items():
            old = f"{k} {value}"
            if k in ("model_end", "scenario_start", "scenario_end"):
                new = f"{k} {pamset.get(k, value)}"
            else:
                new = f"{k} {pamset.get(k, float(value)):.4}"
            filedata = filedata.replace(old, new)

        with open(
            os.path.join(filedir, "inputfiles", "pam_current.scm"),
            "w",
            encoding="ascii",
        ) as scfile:
            scfile.write(filedata)



import re

import pytest

from openscm_runner import run


def test_run_out_config_conflict_error():
    error_msg = re.escape(
        "Found model(s) in `out_config` which are not in "
        "`climate_models_cfgs`: {'another model'}"
    )
    with pytest.raises(NotImplementedError):
        with pytest.warns(UserWarning, match=error_msg):
            run(
                climate_models_cfgs={"model_a": ["config list"]},
                scenarios="not used",
                out_config={"another model": ("hi",)},
            )


def test_run_out_config_type_error():
    error_msg = re.escape(
        "`out_config` values must be tuples, this isn't the case for "
        "climate_model: 'model_a'"
    )
    with pytest.raises(TypeError, match=error_msg):
        run(
            climate_models_cfgs={"model_a": ["config list"]},
            scenarios="not used",
            out_config={"model_a": "hi"},
        )



from tqdm.autonotebook import tqdm

from openscm_runner.progress import _default_tqdm_params, progress


def test_progress(monkeypatch):
    items = range(10)

    iterable = progress(items)

    assert isinstance(iterable, tqdm)
    assert iterable.mininterval == 5
    assert iterable.unit == "it"

    monkeypatch.setitem(_default_tqdm_params, "mininterval", 2)

    iterable = progress(items)
    assert iterable.mininterval == 2


def test_progress_disable(monkeypatch):
    items = range(10)

    monkeypatch.setitem(_default_tqdm_params, "disable", True)

    iterable = progress(items)
    assert iterable.disable



from abc import ABC, abstractmethod

import numpy.testing as npt
from scmdata import ScmRun

import openscm_runner.testing
from openscm_runner import run


class _AdapterTester(ABC):
    """
    Base class for testing adapters.

    All adapters should have a test class which sub-classes this class and
    implements all abstract methods.
    """

    """float: relative tolerance for numeric comparisons"""
    _rtol = 1e-5

    """
    tuple[str]: variable names which are expected to be used by all adapters

    This could be replaced with the RCMIP variables in future (bigger job)
    """
    _common_variables = (
        "Surface Air Temperature Change",
        "Surface Air Ocean Blended Temperature Change",
        "Effective Radiative Forcing",
        "Effective Radiative Forcing|Anthropogenic",
        "Effective Radiative Forcing|Aerosols",
        "Effective Radiative Forcing|Aerosols|Direct Effect",
        "Effective Radiative Forcing|Aerosols|Direct Effect|BC",
        "Effective Radiative Forcing|Aerosols|Direct Effect|BC|MAGICC Fossil and Industrial",
        "Effective Radiative Forcing|Aerosols|Direct Effect|BC|MAGICC AFOLU",
        "Effective Radiative Forcing|Aerosols|Direct Effect|OC",
        "Effective Radiative Forcing|Aerosols|Direct Effect|OC|MAGICC Fossil and Industrial",
        "Effective Radiative Forcing|Aerosols|Direct Effect|OC|MAGICC AFOLU",
        "Effective Radiative Forcing|Aerosols|Direct Effect|SOx",
        "Effective Radiative Forcing|Aerosols|Direct Effect|SOx|MAGICC Fossil and Industrial",
        "Effective Radiative Forcing|Aerosols|Direct Effect|SOx|MAGICC AFOLU",
        "Effective Radiative Forcing|Aerosols|Indirect Effect",
        "Effective Radiative Forcing|Greenhouse Gases",
        "Effective Radiative Forcing|CO2",
        "Effective Radiative Forcing|CH4",
        "Effective Radiative Forcing|N2O",
        "Effective Radiative Forcing|F-Gases",
        "Effective Radiative Forcing|HFC125",
        "Effective Radiative Forcing|HFC134a",
        "Effective Radiative Forcing|HFC143a",
        "Effective Radiative Forcing|HFC227ea",
        "Effective Radiative Forcing|HFC23",
        "Effective Radiative Forcing|HFC245fa",
        "Effective Radiative Forcing|HFC32",
        "Effective Radiative Forcing|HFC4310mee",
        "Effective Radiative Forcing|CF4",
        "Effective Radiative Forcing|C6F14",
        "Effective Radiative Forcing|C2F6",
        "Effective Radiative Forcing|SF6",
        "Heat Uptake",
        "Heat Uptake|Ocean",
        "Atmospheric Concentrations|CO2",
        "Atmospheric Concentrations|CH4",
        "Atmospheric Concentrations|N2O",
        "Net Atmosphere to Land Flux|CO2",
        "Net Atmosphere to Ocean Flux|CO2",
    )

    def _check_res(self, exp, check_val, raise_error):
        try:
            npt.assert_allclose(exp, check_val, rtol=self._rtol)
        except AssertionError:
            if raise_error:
                raise

            print("exp: {}, check_val: {}".format(exp, check_val))

    def _check_output(self, res, expected_output_file, update):
        openscm_runner.testing._check_output(
            res, expected_output_file, self._rtol, update
        )

    @staticmethod
    def _check_heat_content_heat_uptake_consistency(res):
        hc_deltas = ScmRun(
            res.filter(variable="Heat Content", region="World")
            .timeseries(time_axis="year")
            .diff(axis="columns")
        )
        hc_deltas["unit"] = "{} / yr".format(hc_deltas.get_unique_meta("unit", True))

        ratio = hc_deltas.divide(
            res.filter(variable="Heat Uptake", region="World"),
            op_cols={"variable": "Heat Content / Heat Uptake"},
        )

        earth_surface_area = 5.100536e14
        npt.assert_allclose(
            earth_surface_area,
            ratio.convert_unit("m^2").timeseries(drop_all_nan_times=True),
        )

    @abstractmethod
    def test_run(self, test_scenarios):
        """
        Test that the model can be run, ideally with more than one configuration
        """
        # pseudo-code
        res = run(
            climate_models_cfgs={
                "climate_model": [
                    {"para1": "value1", "para2": 1.1},
                    {"para1": "value3", "para2": 1.2},
                    {"para1": "value5", "para2": 1.3},
                ],
            },
            scenarios=test_scenarios.filter(scenario=["ssp126", "ssp245", "ssp370"]),
            output_variables=("output", "var", "list",),
            out_config=None,
        )

        assert isinstance(res, ScmRun)
        assert res["run_id"].min() == 0
        assert res["run_id"].max() == 8

        assert res.get_unique_meta("climate_model", no_duplicates=True) == "model_name"

        assert set(res.get_unique_meta("variable")) == set(
            ["expected", "output", "variables"]
        )

        # output value checks e.g.
        npt.assert_allclose(
            2.31,
            res.filter(
                variable="Surface Air Temperature Change",
                region="World",
                year=2100,
                scenario="ssp126",
            ).values.max(),
            rtol=self._rtol,
        )

    @abstractmethod
    def test_variable_naming(self, test_scenarios):
        """
        Test that variable naming is implemented as expecteds
        """
        # pseudo-code

        # a model might not report all outputs
        missing_variables = (
            "Net Atmosphere to Ocean Flux|CO2",
            "Net Atmosphere to Land Flux|CO2",
        )
        common_variables = [
            c for c in self._common_variables if c not in missing_variables
        ]

        # run the model and request all common variables the model reportss
        res = run(
            climate_models_cfgs={"climate_model": ({"para1": 3.43},)},
            scenarios=test_scenarios.filter(scenario="ssp126"),
            output_variables=common_variables,
        )

        # check that all expected outputs were returned
        missing_vars = set(common_variables) - set(res["variable"])
        if missing_vars:
            raise AssertionError(missing_vars)



"""
Test that all of our modules can be imported

Thanks https://stackoverflow.com/a/25562415/10473080
"""
import importlib
import os.path
import pkgutil

import openscm_runner
import openscm_runner.adapters.fair_adapter.fair_adapter


def import_submodules(package_name):
    package = importlib.import_module(package_name)

    for _, name, is_pkg in pkgutil.walk_packages(package.__path__):
        full_name = package.__name__ + "." + name
        print(full_name)
        importlib.import_module(full_name)
        if is_pkg:
            import_submodules(full_name)


import_submodules("openscm_runner")

# make sure csvs etc. are included
openscm_runner_root = os.path.dirname(openscm_runner.__file__)
assert os.path.isfile(
    os.path.join(
        openscm_runner_root, "adapters/fair_adapter/natural-emissions-and-forcing.csv"
    )
)
assert os.path.isfile(
    os.path.join(
        openscm_runner_root,
        "adapters/ciceroscm_adapter/utils_templates/gases_v1RCMIP.txt",
    )
)
assert os.path.isfile(
    os.path.join(
        openscm_runner_root,
        "adapters/ciceroscm_adapter/utils_templates/run_dir/scm_vCH4fb",
    )
)
assert os.path.isfile(
    os.path.join(
        openscm_runner_root,
        "adapters/ciceroscm_adapter/utils_templates/run_dir/input_RF/RFSUN/solar_IPCC.txt",
    )
)

print(openscm_runner.__version__)



import os.path

import pymagicc.io
import pytest
from base import _AdapterTester
from scmdata import ScmRun

from openscm_runner import run
from openscm_runner.adapters import MAGICC7
from openscm_runner.utils import calculate_quantiles


@pytest.mark.magicc
class TestMagicc7Adapter(_AdapterTester):
    def test_run(
        self, test_scenarios, test_data_dir, update_expected_values,
    ):
        expected_output_file = os.path.join(
            test_data_dir,
            "expected-integration-output",
            "expected_magicc7_test_run_output.json",
        )

        res = run(
            climate_models_cfgs={
                "MAGICC7": [
                    {
                        "core_climatesensitivity": 3,
                        "rf_soxi_dir_wm2": -0.2,
                        "out_temperature": 1,
                        "out_forcing": 1,
                        "out_dynamic_vars": [
                            "DAT_AEROSOL_ERF",
                            "DAT_HEATCONTENT_AGGREG_TOTAL",
                            "DAT_CO2_AIR2LAND_FLUX",
                        ],
                        "out_ascii_binary": "BINARY",
                        "out_binary_format": 2,
                    },
                    {
                        "core_climatesensitivity": 2,
                        "rf_soxi_dir_wm2": -0.1,
                        "out_temperature": 1,
                        "out_forcing": 1,
                        "out_ascii_binary": "BINARY",
                        "out_binary_format": 2,
                    },
                    {
                        "core_climatesensitivity": 5,
                        "rf_soxi_dir_wm2": -0.35,
                        "out_temperature": 1,
                        "out_forcing": 1,
                        "out_ascii_binary": "BINARY",
                        "out_binary_format": 2,
                    },
                ],
            },
            scenarios=test_scenarios.filter(scenario=["ssp126", "ssp245", "ssp370"]),
            output_variables=(
                "Surface Air Temperature Change",
                "Effective Radiative Forcing",
                "Effective Radiative Forcing|Aerosols",
                "Effective Radiative Forcing|CO2",
                "Heat Content",
                "Heat Content|Ocean",
                "Heat Uptake",
                "Heat Uptake|Ocean",
                "Net Atmosphere to Land Flux|CO2",
            ),
        )

        assert isinstance(res, ScmRun)
        assert res["run_id"].min() == 0
        assert res["run_id"].max() == 8
        assert res.get_unique_meta(
            "climate_model", no_duplicates=True
        ) == "MAGICC{}".format(MAGICC7.get_version())
        assert set(res.get_unique_meta("variable")) == set(
            [
                "Surface Air Temperature Change",
                "Effective Radiative Forcing",
                "Effective Radiative Forcing|Aerosols",
                "Effective Radiative Forcing|CO2",
                "Heat Content",
                "Heat Content|Ocean",
                "Heat Uptake",
                "Heat Uptake|Ocean",
                "Net Atmosphere to Land Flux|CO2",
            ]
        )

        # check we can also calcluate quantiles
        assert "run_id" in res.meta
        quantiles = calculate_quantiles(res, [0, 0.05, 0.17, 0.5, 0.83, 0.95, 1])
        assert "run_id" not in quantiles.meta

        # a problem for another day...
        # self._check_heat_content_heat_uptake_consistency(res)

        self._check_output(res, expected_output_file, update_expected_values)

    def test_variable_naming(self, test_scenarios):
        common_variables = self._common_variables
        res = run(
            climate_models_cfgs={"MAGICC7": ({"core_climatesensitivity": 3},)},
            scenarios=test_scenarios.filter(scenario="ssp126"),
            output_variables=common_variables,
        )

        missing_vars = set(common_variables) - set(res["variable"])
        if missing_vars:
            raise AssertionError(missing_vars)


@pytest.mark.magicc
def test_write_scen_files_and_make_full_cfgs(test_scenarios):
    adapter = MAGICC7()
    test_scenarios_magiccdf = pymagicc.io.MAGICCData(test_scenarios)
    res = adapter._write_scen_files_and_make_full_cfgs(
        test_scenarios_magiccdf,
        [
            {
                "file_emisscen_3": "overwritten by adapter.magicc_scenario_setup",
                "other_cfg": 12,
            }
        ],
    )

    for (model, scenario), _ in test_scenarios_magiccdf.meta.groupby(
        ["model", "scenario"]
    ):
        scen_file_name = (
            "{}_{}.SCEN7".format(scenario, model)
            .upper()
            .replace("/", "-")
            .replace("\\", "-")
            .replace(" ", "-")
        )

        scenario_cfg = [v for v in res if v["file_emisscen"] == scen_file_name]

        assert len(scenario_cfg) == 1
        scenario_cfg = scenario_cfg[0]
        assert scenario_cfg["other_cfg"] == 12
        assert scenario_cfg["model"] == model
        assert scenario_cfg["scenario"] == scenario
        for i in range(2, 9):
            scen_flag_val = scenario_cfg["file_emisscen_{}".format(i)]

            assert scen_flag_val == "NONE"


@pytest.mark.magicc
@pytest.mark.parametrize(
    "out_config",
    (
        ("core_climatesensitivity", "rf_total_runmodus"),
        ("core_climatesensitivity",),
        ("rf_total_runmodus",),
    ),
)
def test_return_config(test_scenarios, out_config):
    core_climatesensitivities = [2, 3]
    rf_total_runmoduses = ["ALL", "CO2"]

    cfgs = []
    for cs in core_climatesensitivities:
        for runmodus in rf_total_runmoduses:
            cfgs.append(
                {
                    "out_dynamic_vars": [
                        "DAT_TOTAL_INCLVOLCANIC_ERF",
                        "DAT_SURFACE_TEMP",
                    ],
                    "core_climatesensitivity": cs,
                    "rf_total_runmodus": runmodus,
                }
            )

    res = run(
        climate_models_cfgs={"MAGICC7": cfgs},
        scenarios=test_scenarios.filter(scenario=["ssp126", "ssp245", "ssp370"]),
        output_variables=(
            "Surface Air Temperature Change",
            "Effective Radiative Forcing",
        ),
        out_config={"MAGICC7": out_config},
    )

    for k in out_config:
        assert k in res.meta.columns
        ssp126 = res.filter(scenario="ssp126")

        # check all the configs were used and check that each scenario
        # has all the configs included in the metadata too
        if k == "core_climatesensitivity":
            assert set(res.get_unique_meta(k)) == set(core_climatesensitivities)
            assert set(ssp126.get_unique_meta(k)) == set(core_climatesensitivities)
        elif k == "rf_total_runmodus":
            assert set(res.get_unique_meta(k)) == set(rf_total_runmoduses)
            assert set(ssp126.get_unique_meta(k)) == set(rf_total_runmoduses)
        else:
            raise NotImplementedError(k)


@pytest.mark.magicc
@pytest.mark.parametrize(
    "cfgs",
    (
        [{"pf_apply": 1, "PF_APPLY": 0}],
        [{"pf_apply": 1}, {"pf_apply": 1, "PF_APPLY": 0}],
    ),
)
def test_return_config_clash_error(test_scenarios, cfgs):
    with pytest.raises(ValueError):
        run(
            climate_models_cfgs={"MAGICC7": cfgs},
            scenarios=test_scenarios.filter(scenario=["ssp126"]),
            output_variables=("Surface Air Temperature Change",),
            out_config={"MAGICC7": ("pf_apply",)},
        )



import os

import numpy as np
import numpy.testing as npt
import pytest
from base import _AdapterTester
from scmdata import ScmRun

from openscm_runner import run
from openscm_runner.adapters import CICEROSCM
from openscm_runner.adapters.ciceroscm_adapter import (
    make_scenario_files,
    write_parameter_files,
)
from openscm_runner.utils import calculate_quantiles

RTOL = 1e-5


class TestCICEROSCMAdapter(_AdapterTester):
    @pytest.mark.ciceroscm
    @pytest.mark.parametrize("shuffle_column_order", (True, False))
    def test_run(
        self,
        test_scenarios,
        test_data_dir,
        update_expected_values,
        shuffle_column_order,
    ):
        expected_output_file = os.path.join(
            test_data_dir,
            "expected-integration-output",
            "expected_ciceroscm_test_run_output.json",
        )

        if shuffle_column_order:
            tmp = test_scenarios.data
            cols = tmp.columns.tolist()
            tmp = tmp[cols[1:] + cols[:1]]
            test_scenarios = ScmRun(test_scenarios)

        res = run(
            scenarios=test_scenarios.filter(scenario=["ssp126", "ssp245", "ssp370"]),
            climate_models_cfgs={
                "CICEROSCM": [
                    {
                        "model_end": 2100,
                        "Index": 30040,
                        "lambda": 0.540,
                        "akapa": 0.341,
                        "cpi": 0.556,
                        "W": 1.897,
                        "rlamdo": 16.618,
                        "beto": 3.225,
                        "mixed": 107.277,
                        "dirso2_forc": -0.457,
                        "indso2_forc": -0.514,
                        "bc_forc": 0.200,
                        "oc_forc": -0.103,
                    },
                    {
                        "model_end": 2100,
                        "Index": 1,
                        "lambda": 0.3925,
                        "akapa": 0.2421,
                        "cpi": 0.3745,
                        "W": 0.8172,
                        "rlamdo": 16.4599,
                        "beto": 4.4369,
                        "mixed": 35.4192,
                        "dirso2_forc": -0.3428,
                        "indso2_forc": -0.3856,
                        "bc_forc": 0.1507,
                        "oc_forc": -0.0776,
                    },
                ]
            },
            output_variables=(
                "Surface Air Temperature Change",
                "Surface Air Ocean Blended Temperature Change",
                "Heat Content|Ocean",
                "Effective Radiative Forcing",
                "Effective Radiative Forcing|Anthropogenic",
                "Effective Radiative Forcing|Aerosols",
                "Effective Radiative Forcing|Greenhouse Gases",
                "Heat Uptake",
                "Atmospheric Concentrations|CO2",
                "Atmospheric Concentrations|CH4",
                "Atmospheric Concentrations|N2O",
                "Emissions|CO2",
                "Emissions|CH4",
                "Emissions|N2O",
            ),
            out_config=None,
        )
        assert isinstance(res, ScmRun)
        assert res["run_id"].min() == 1
        assert res["run_id"].max() == 30040
        assert res.get_unique_meta("climate_model", no_duplicates=True) == "CICERO-SCM"

        assert set(res.get_unique_meta("variable")) == set(
            [
                "Surface Air Temperature Change",
                "Surface Air Ocean Blended Temperature Change",
                "Heat Content|Ocean",
                "Effective Radiative Forcing",
                "Effective Radiative Forcing|Anthropogenic",
                "Effective Radiative Forcing|Aerosols",
                "Effective Radiative Forcing|Greenhouse Gases",
                "Heat Uptake",
                "Atmospheric Concentrations|CO2",
                "Atmospheric Concentrations|CH4",
                "Atmospheric Concentrations|N2O",
                "Emissions|CO2",
                "Emissions|N2O",
                "Emissions|CH4",
            ]
        )

        # check we can also calcluate quantiles
        assert "run_id" in res.meta
        quantiles = calculate_quantiles(res, [0, 0.05, 0.17, 0.5, 0.83, 0.95, 1])
        assert "run_id" not in quantiles.meta

        assert (
            res.filter(variable="Atmospheric Concentrations|CO2").get_unique_meta(
                "unit", True
            )
            == "ppm"
        )
        assert (
            res.filter(variable="Atmospheric Concentrations|CH4").get_unique_meta(
                "unit", True
            )
            == "ppb"
        )
        assert (
            res.filter(variable="Atmospheric Concentrations|N2O").get_unique_meta(
                "unit", True
            )
            == "ppb"
        )
        assert (
            res.filter(variable="Emissions|CO2").get_unique_meta("unit", True)
            == "PgC / yr"
        )
        assert (
            res.filter(variable="Emissions|CH4").get_unique_meta("unit", True)
            == "TgCH4 / yr"
        )
        assert (
            res.filter(variable="Emissions|N2O").get_unique_meta("unit", True)
            == "TgN2ON / yr"
        )

        # check that emissions were passed through correctly
        for (scen, variable, unit, exp_val) in (
            ("ssp126", "Emissions|CO2", "PgC/yr", -2.3503),
            ("ssp370", "Emissions|CO2", "PgC/yr", 22.562),
            ("ssp126", "Emissions|CH4", "TgCH4/yr", 122.195),
            ("ssp370", "Emissions|CH4", "TgCH4/yr", 777.732),
            ("ssp126", "Emissions|N2O", "TgN2ON/yr", 5.318),
            ("ssp370", "Emissions|N2O", "TgN2ON/yr", 13.144),
        ):
            res_scen_2100_emms = res.filter(
                variable=variable, year=2100, scenario=scen
            ).convert_unit(unit)
            if res_scen_2100_emms.empty:
                raise AssertionError("No {} data for {}".format(variable, scen))

            npt.assert_allclose(
                res_scen_2100_emms.values, exp_val, rtol=1e-4,
            )
        for (scen, variable, unit, exp_val14, exp_val16) in (
            ("ssp126", "Emissions|CH4", "TgCH4/yr", 387.874, 379.956),
            ("ssp370", "Emissions|CH4", "TgCH4/yr", 387.874, 394.149),
            ("ssp126", "Emissions|N2O", "TgN2ON/yr", 6.911, 6.858),
            ("ssp370", "Emissions|N2O", "TgN2ON/yr", 6.911, 7.0477),
        ):
            res_scen_2014_emms = res.filter(
                variable=variable, year=2014, scenario=scen
            ).convert_unit(unit)
            if res_scen_2014_emms.empty:
                raise AssertionError("No {} data for {}".format(variable, scen))

            res_scen_2016_emms = res.filter(
                variable=variable, year=2016, scenario=scen
            ).convert_unit(unit)
            if res_scen_2016_emms.empty:
                raise AssertionError("No {} data for {}".format(variable, scen))

            npt.assert_allclose(
                res_scen_2014_emms.values, exp_val14, rtol=1e-4,
            )
            npt.assert_allclose(
                res_scen_2016_emms.values, exp_val16, rtol=1e-4,
            )
        for (scen, variable) in (
            ("ssp126", "Effective Radiative Forcing|Aerosols"),
            ("ssp370", "Effective Radiative Forcing|Aerosols"),
        ):
            res_scen_2015_emms = res.filter(variable=variable, year=2015, scenario=scen)
            if res_scen_2015_emms.empty:
                raise AssertionError("No CO2 emissions data for {}".format(scen))

            assert not np.equal(res_scen_2015_emms.values, 0).all()

        ssp245_ghg_erf_2015 = res.filter(
            variable="Effective Radiative Forcing|Greenhouse Gases",
            year=2015,
            scenario="ssp245",
            run_id=1,
        )
        ssp245_ghg_erf_2014 = res.filter(
            variable="Effective Radiative Forcing|Greenhouse Gases",
            year=2014,
            scenario="ssp245",
            run_id=1,
        )
        # check that jump in GHG ERF isn't there
        assert (
            ssp245_ghg_erf_2015.values.squeeze() - ssp245_ghg_erf_2014.values.squeeze()
        ) < 0.1
        ssp245_ch4_conc_2015 = res.filter(
            variable="Atmospheric Concentrations|CH4",
            year=2015,
            scenario="ssp245",
            run_id=1,
        )
        ssp245_ch4_conc_2014 = res.filter(
            variable="Atmospheric Concentrations|CH4",
            year=2014,
            scenario="ssp245",
            run_id=1,
        )
        # ch
        # check that jump in GHG ERF isn't there
        assert (
            ssp245_ch4_conc_2014.values.squeeze()
            - ssp245_ch4_conc_2015.values.squeeze()
        ) < 0.1

        self._check_output(res, expected_output_file, update_expected_values)

    @pytest.mark.ciceroscm
    def test_variable_naming(self, test_scenarios):
        missing_from_ciceroscm = (
            "Effective Radiative Forcing|Aerosols|Direct Effect|BC|MAGICC AFOLU",
            "Effective Radiative Forcing|Aerosols|Direct Effect|BC|MAGICC Fossil and Industrial",
            "Effective Radiative Forcing|Aerosols|Direct Effect|OC|MAGICC AFOLU",
            "Effective Radiative Forcing|Aerosols|Direct Effect|OC|MAGICC Fossil and Industrial",
            "Effective Radiative Forcing|Aerosols|Direct Effect|SOx|MAGICC AFOLU",
            "Effective Radiative Forcing|Aerosols|Direct Effect|SOx|MAGICC Fossil and Industrial",
            "Heat Uptake|Ocean",
            "Net Atmosphere to Land Flux|CO2",
            "Net Atmosphere to Ocean Flux|CO2",
        )
        common_variables = [
            c for c in self._common_variables if c not in missing_from_ciceroscm
        ]
        res = run(
            climate_models_cfgs={"CICEROSCM": ({"lambda": 0.540},)},
            scenarios=test_scenarios.filter(scenario="ssp126"),
            output_variables=common_variables,
        )

        missing_vars = set(common_variables) - set(res["variable"])
        if missing_vars:
            raise AssertionError(missing_vars)

    @pytest.mark.ciceroscm
    def test_w_out_config(self, test_scenarios):
        with pytest.raises(NotImplementedError):
            run(
                scenarios=test_scenarios.filter(scenario=["ssp126"]),
                climate_models_cfgs={
                    "CiceroSCM": [
                        {
                            "model_end": 2100,
                            "Index": 30040,
                            "lambda": 0.540,
                            "akapa": 0.341,
                            "cpi": 0.556,
                            "W": 1.897,
                            "rlamdo": 16.618,
                            "beto": 3.225,
                            "mixed": 107.277,
                            "dirso2_forc": -0.457,
                            "indso2_forc": -0.514,
                            "bc_forc": 0.200,
                            "oc_forc": -0.103,
                        },
                    ]
                },
                output_variables=("Surface Air Temperature Change",),
                out_config={"CiceroSCM": ("With ECS",)},
            )

    def test_make_scenario_files(self, test_scenarios):
        npt.assert_allclose(
            3.0 / 11.0 * 1000.0,
            make_scenario_files._unit_conv_factor("Gg CO2/yr", "Mg C/ yr"),
        )
        npt.assert_allclose(
            28 / 44 / 1.0e12,
            make_scenario_files._unit_conv_factor("kg N2O / yr", "Pg N2ON / yr"),
        )
        npt.assert_allclose(
            14 / 46 / 1.0e12,
            make_scenario_files._unit_conv_factor("kt NOx / yr", "Pt N / yr"),
        )
        npt.assert_allclose(
            0.5, make_scenario_files._unit_conv_factor("Tg SO2 / yr", "Tg S / yr"),
        )
        npt.assert_allclose(
            1.0,
            make_scenario_files._unit_conv_factor("Gg  Halon1211 / yr", "GgH1211 / yr"),
        )
        npt.assert_allclose(
            1.0,
            make_scenario_files._unit_conv_factor("Gg  Halon2402 / yr", "GgH2402 / yr"),
        )

    @pytest.mark.parametrize(
        "input,exp",
        (
            ("folder", ("folder",)),
            (os.path.join("folder", "subfolder"), ("folder", "subfolder")),
        ),
    )
    def test_write_parameter_files(self, input, exp):
        assert write_parameter_files.splitall(input) == exp

    @pytest.mark.ciceroscm
    @pytest.mark.parametrize(
        "name",
        (
            "some super super super super super super super super super super long scenario name which should not explode",
            "another name with some special | . characters which should also work",
            # any other names you want to test
        ),
    )
    def test_run_long_scenario_name(
        self, name, test_scenarios,
    ):
        starting_scenario = test_scenarios.filter(scenario="ssp126").rename(
            {"scenario": {"ssp126": name}}
        )

        res = run(
            scenarios=starting_scenario,
            climate_models_cfgs={"CICEROSCM": [{}]},
            output_variables=("Surface Air Temperature Change",),
        )

        assert res.get_unique_meta("scenario", True) == name
        assert (1.4 < res.filter(year=2100).values < 1.6).all()
        # any other tests you want to add


@pytest.mark.ciceroscm
def test_get_version():
    assert CICEROSCM.get_version() == "v2019vCH4"



"""
MAGICC7 adapter
"""
import logging
import os
from subprocess import check_output  # nosec

from scmdata import ScmRun, run_append

from ...progress import progress
from ...settings import config
from ..base import _Adapter
from ._compat import pymagicc
from ._run_magicc_parallel import run_magicc_parallel

LOGGER = logging.getLogger(__name__)


_VARIABLE_MAP = {
    "Surface Air Temperature Change": "Surface Temperature",
    "Atmospheric Concentrations|HFC4310mee": "Atmospheric Concentrations|HFC4310",
    "Radiative Forcing|HFC4310mee": "Radiative Forcing|HFC4310",
    "Effective Radiative Forcing|HFC4310mee": "Effective Radiative Forcing|HFC4310",
}
"""
dict[str: str] : Mapping from openscm_runner names to pymagicc names

Should only be used as an emergency escape, if changes are needed
they should really be made in pymagicc.
"""


def _convert_to_pymagicc_var(in_var):
    """
    Convert an OpenSCM-Runner name to a Pymagicc name
    """
    if in_var in _VARIABLE_MAP:
        return _VARIABLE_MAP[in_var]

    out = pymagicc.definitions.convert_magicc7_to_openscm_variables(in_var)

    return out


class MAGICC7(_Adapter):
    """
    Adapter for running MAGICC7

    The adapter overwrites all of MAGICC7's emissions flags so that only
    emissions passed from the user are used.
    """

    def __init__(self):
        """
        Initialise the MAGICC7 adapter
        """
        if pymagicc is None:
            raise ImportError(
                "pymagicc is not installed. Run 'conda install pymagicc' or 'pip install pymagicc'"
            )
        super().__init__()
        self.magicc_scenario_setup = {
            "file_emisscen_2": "NONE",
            "file_emisscen_3": "NONE",
            "file_emisscen_4": "NONE",
            "file_emisscen_5": "NONE",
            "file_emisscen_6": "NONE",
            "file_emisscen_7": "NONE",
            "file_emisscen_8": "NONE",
        }
        """dict: MAGICC base scenario setup"""

    def _init_model(self):  # pylint:disable=arguments-differ
        pass

    @staticmethod
    def _convert_to_magicc_units(scenarios):
        magicc_scmdf = pymagicc.io.MAGICCData(scenarios)
        emms_units = pymagicc.definitions.MAGICC7_EMISSIONS_UNITS
        emms_units["openscm_variable"] = emms_units["magicc_variable"].apply(
            lambda x: pymagicc.definitions.convert_magicc7_to_openscm_variables(
                f"{x}_EMIS"
            )
        )
        emms_units = emms_units.set_index("openscm_variable")
        for variable in magicc_scmdf["variable"].unique():
            magicc_unit = emms_units.loc[variable, "emissions_unit"]
            if "NOx" in variable:
                context = "NOx_conversions"
            elif "NH3" in variable:
                context = "NH3_conversions"
            else:
                context = None
            magicc_scmdf = magicc_scmdf.convert_unit(
                magicc_unit, variable=variable, context=context,
            )

        return magicc_scmdf

    def _run(self, scenarios, cfgs, output_variables, output_config):
        # TODO: add use of historical data properly  # pylint:disable=fixme
        LOGGER.warning("Historical data has not been checked")

        magicc_df = scenarios.timeseries().reset_index()
        magicc_df["variable"] = magicc_df["variable"].apply(
            lambda x: x.replace("Sulfur", "SOx")
            .replace("HFC4310mee", "HFC4310")
            .replace("VOC", "NMVOC")
        )

        magicc_scmdf = self._convert_to_magicc_units(magicc_df)
        full_cfgs = self._write_scen_files_and_make_full_cfgs(magicc_scmdf, cfgs)

        pymagicc_vars = [_convert_to_pymagicc_var(v) for v in output_variables]
        res = run_magicc_parallel(full_cfgs, pymagicc_vars, output_config)

        LOGGER.debug("Dropping todo metadata")
        res = res.drop_meta("todo")
        res["climate_model"] = f"MAGICC{self.get_version()}"

        res = self._fix_pint_incompatible_units(res)
        LOGGER.debug("Mapping variables to OpenSCM conventions")
        inverse_map = {v: k for k, v in _VARIABLE_MAP.items()}
        res["variable"] = res["variable"].apply(
            lambda x: inverse_map[x] if x in inverse_map else x
        )

        res = ScmRun(res)

        return res

    @staticmethod
    def _fix_pint_incompatible_units(inp):
        out = inp

        conversions = (("10^22 J", 10, "ZJ"),)
        for odd_unit, conv_factor, new_unit in conversions:
            if odd_unit in inp.get_unique_meta("unit"):
                LOGGER.debug(
                    "Converting %s to %s with a conversion factor of %f",
                    odd_unit,
                    new_unit,
                    conv_factor,
                )
                rest_ts = inp.filter(unit=odd_unit, keep=False)
                odd_unit_ts = inp.filter(unit=odd_unit)
                odd_unit_ts *= conv_factor
                odd_unit_ts["unit"] = new_unit
                out = run_append([rest_ts, odd_unit_ts])

        return out

    def _write_scen_files_and_make_full_cfgs(self, scenarios, cfgs):
        full_cfgs = []
        run_id_block = 0

        for (scenario, model), smdf in progress(
            scenarios.timeseries().groupby(["scenario", "model"]),
            desc="Writing SCEN7 files",
        ):
            writer = pymagicc.io.MAGICCData(smdf)
            writer["todo"] = "SET"
            writer.metadata = {
                "header": f"SCEN7 file written by openscm_runner for the {scenario} scenario"
            }
            scen_file_name = (
                f"{scenario}_{model}.SCEN7".upper()
                .replace("/", "-")
                .replace("\\", "-")
                .replace(" ", "-")
            )
            writer.write(
                os.path.join(self._run_dir(), scen_file_name),
                magicc_version=self.get_version()[1],
            )

            scenario_cfg = [
                {
                    "scenario": scenario,
                    "model": model,
                    "file_emisscen": scen_file_name,
                    "run_id": i + run_id_block,
                    **cfg,
                    **self.magicc_scenario_setup,
                }
                for i, cfg in enumerate(cfgs)
            ]
            run_id_block += len(scenario_cfg)

            full_cfgs += scenario_cfg

        exp_shape = scenarios.meta[["scenario", "model"]].drop_duplicates().shape[
            0
        ] * len(cfgs)
        if len(full_cfgs) != exp_shape:
            raise AssertionError(f"Expected {exp_shape} configs got {len(full_cfgs)}")
        return full_cfgs

    @classmethod
    def get_version(cls):
        """
        Get the MAGICC7 version being used by this adapter

        Returns
        -------
        str
            The MAGICC7 version id
        """
        return check_output([cls._executable(), "--version"]).decode("utf-8").strip()

    @classmethod
    def _executable(cls):
        return config["MAGICC_EXECUTABLE_7"]

    @classmethod
    def _run_dir(cls):
        return os.path.join(os.path.dirname(cls._executable()), "..", "run")



"""
Module that reads in CICERO-SCM results
and returns data to append to SCMRun
"""
import os
import re

import numpy as np
import pandas as pd


def get_data_from_conc_file(folder, variable):
    """
    Get data from concentration files
    """
    df_temp = pd.read_csv(os.path.join(folder, "temp_conc.txt"), delimiter=r"\s+")
    years = df_temp.Year[:]
    timeseries = df_temp[variable].to_numpy()  # pylint:disable=unsubscriptable-object
    return years, timeseries


def get_data_from_em_file(folder, variable):
    """
    Get data from emissions files
    """
    df_temp = pd.read_csv(os.path.join(folder, "temp_em.txt"), delimiter=r"\s+")
    years = df_temp.Year[:]
    timeseries = df_temp[variable].to_numpy()  # pylint:disable=unsubscriptable-object
    return years, timeseries


def get_data_from_temp_file(folder, variable):
    """
    Get data from temperature files
    """
    df_temp = pd.read_csv(os.path.join(folder, "temp_temp.txt"), delimiter=r"\s+")
    years = df_temp.Year[:]
    timeseries = df_temp[variable].to_numpy()  # pylint:disable=unsubscriptable-object
    return years, timeseries


def get_data_from_ohc_file(folder, variable):
    """
    Get data from ocean heat content files
    """
    df_temp = pd.read_csv(os.path.join(folder, "temp_ohc.txt"), delimiter=r"\s+")
    years = df_temp.Year[:]
    # Units are 10^22J and output should be 10^21J = ZJ
    conv_factor = 10.0
    timeseries = (
        df_temp[variable].to_numpy()  # pylint:disable=unsubscriptable-object
        * conv_factor
    )
    return years, timeseries


def get_data_from_rib_file(folder, variable):
    """
    Get data from rib files
    """
    df_temp = pd.read_csv(os.path.join(folder, "temp_rib.txt"), delimiter=r"\s+")
    years = df_temp.Year[:]
    timeseries = df_temp[variable].to_numpy()  # pylint:disable=unsubscriptable-object
    return years, timeseries


def convert_cicero_unit(cicero_unit):
    """
    Convert cicero unit convention for pint
    """
    return f"{cicero_unit.replace('_', '')} / yr"


class CSCMREADER:
    """
    Class to read CICERO-SCM output data
    """

    def __init__(self, odir):
        self.odir = odir
        self.variable_dict = {
            "Surface Air Temperature Change": "dT_glob_air",
            # GMST
            "Surface Air Ocean Blended Temperature Change": "dT_glob",
            # ERFs
            "Effective Radiative Forcing": "Total_forcing+sunvolc",
            "Effective Radiative Forcing|Anthropogenic": "Total_forcing",
            "Effective Radiative Forcing|Aerosols": "Aerosols",
            "Effective Radiative Forcing|Aerosols|Direct Effect": "Aerosols|Direct Effect",
            "Effective Radiative Forcing|Aerosols|Direct Effect|BC": "BC",
            "Effective Radiative Forcing|Aerosols|Direct Effect|OC": "OC",
            "Effective Radiative Forcing|Aerosols|Direct Effect|SOx": "SO2",
            "Effective Radiative Forcing|Aerosols|Indirect Effect": "SO4_IND",
            "Effective Radiative Forcing|Greenhouse Gases": "GHG",
            "Effective Radiative Forcing|F-Gases": "Fgas",
            "Effective Radiative Forcing|HFC125": "HFC125",
            "Effective Radiative Forcing|HFC134a": "HFC134a",
            "Effective Radiative Forcing|HFC143a": "HFC143a",
            "Effective Radiative Forcing|HFC227ea": "HFC227ea",
            "Effective Radiative Forcing|HFC23": "HFC23",
            "Effective Radiative Forcing|HFC245fa": "HFC245fa",
            "Effective Radiative Forcing|HFC32": "HFC32",
            "Effective Radiative Forcing|HFC4310mee": "HFC4310mee",
            "Effective Radiative Forcing|CF4": "CF4",
            "Effective Radiative Forcing|C6F14": "C6F14",
            "Effective Radiative Forcing|C2F6": "C2F6",
            "Effective Radiative Forcing|SF6": "SF6",
            "Effective Radiative Forcing|CO2": "CO2",
            "Effective Radiative Forcing|CH4": "CH4",
            "Effective Radiative Forcing|N2O": "N2O",
            "Emissions|CO2": "CO2",
            "Emissions|CH4": "CH4",
            "Emissions|N2O": "N2O",
            # Heat uptake
            "Heat Uptake": "RIB_glob",
            "Heat Content|Ocean": "OHCTOT",
            # concentrations
            "Atmospheric Concentrations|CO2": "CO2",
            "Atmospheric Concentrations|CH4": "CH4",
            "Atmospheric Concentrations|N2O": "N2O",
        }
        self.rib_list = "RIB_glob"
        self.temp_list = (
            "dT_glob",
            "dT_glob_air",
            "dT_glob_sea",
            "dSL(m)",
            "dSL_thermal(m)",
            "dSL_ice(m)",
        )
        self.ohc_list = "OHCTOT"
        self.volc_series = self.read_volc_series()
        self.sun_series = self.read_sun_series()

    def read_volc_series(self):
        """
        Read volcanic forcing timeseries from input
        """
        vfile = os.path.join(self.odir, "input_RF", "RFVOLC", "meanVOLCmnd_ipcc_NH.txt")
        volc_series = pd.read_csv(
            vfile, sep="\t", usecols=[0], index_col=False, header=None
        )
        volc_series.index = range(1750, 2501)
        return volc_series

    def read_sun_series(self):
        """
        Read solar forcing timeseries from input
        """
        sfile = os.path.join(self.odir, "input_RF", "RFSUN", "solar_IPCC.txt")
        sun_series = pd.read_csv(sfile, sep="\t", index_col=False, header=None)
        sun_series.index = range(1750, 2501)
        return sun_series

    def read_variable_timeseries(self, scenario, variable, sfilewriter):
        """
        Read variable timeseries
        Connecting up to correct file type to get the data
        """
        folder = os.path.join(
            self.odir, re.sub("[^a-zA-Z0-9_-]", "", scenario)[:50], "outputfiles"
        )
        if variable not in self.variable_dict:
            return (
                pd.Series([], dtype="float64"),
                pd.Series([], dtype="float64"),
                "NoUnit",
            )
        if "Concentration" in variable:
            years, timeseries = get_data_from_conc_file(
                folder, self.variable_dict[variable]
            )
            unit = sfilewriter.concunits[
                sfilewriter.components.index(self.variable_dict[variable])
            ]
        elif "Emissions" in variable:
            years, timeseries = get_data_from_em_file(
                folder, self.variable_dict[variable]
            )
            unit = sfilewriter.units[
                sfilewriter.components.index(self.variable_dict[variable])
            ]
        elif "Forcing" in variable:
            years, timeseries = self.get_data_from_forc_file(
                folder, self.variable_dict[variable]
            )
            unit = "W/m^2"
        elif self.variable_dict[variable] in self.temp_list:
            years, timeseries = get_data_from_temp_file(
                folder, self.variable_dict[variable]
            )
            unit = "K"
        elif self.variable_dict[variable] in self.rib_list:
            years, timeseries = get_data_from_rib_file(
                folder, self.variable_dict[variable]
            )
            unit = "W/m^2"

        elif self.variable_dict[variable] in self.ohc_list:
            years, timeseries = get_data_from_ohc_file(
                folder, self.variable_dict[variable]
            )
            unit = "ZJ"

        return years, timeseries, unit

    def get_volc_forcing(self, startyear, endyear):
        """
        Return volcanic forcing time series from startyear up to and including endyear
        """
        return self.volc_series.iloc[
            startyear - 1750 : endyear - 1750 + 1
        ].values.flatten()

    def get_sun_forcing(self, startyear, endyear):
        """
        Return volcanic forcing time series from startyear up to and including endyear
        """
        return self.sun_series.iloc[
            startyear - 1750 : endyear - 1750 + 1
        ].values.flatten()

    def get_data_from_forc_file(self, folder, variable):
        """
        Get data from forcing files
        """
        forc_sums = ["Aerosols", "Aerosols|Direct Effect"]
        fgas_list = [
            "CFC-11",
            "CFC-12",
            "CFC-113",
            "CFC-114",
            "CFC-115",
            "CH3Br",
            "CCl4",
            "CH3CCl3",
            "HCFC-22",
            "HCFC-141b",
            "HCFC-123",
            "HCFC-142b",
            "H-1211",
            "H-1301",
            "H-2402",
            "HFC125",
            "HFC134a",
            "HFC143a",
            "HFC227ea",
            "HFC23",
            "HFC245fa",
            "HFC32",
            "HFC4310mee",
            "C2F6",
            "C6F14",
            "CF4",
            "SF6",
        ]
        ghg_not_fgas = ["CO2", "CH4", "N2O", "TROP_O3", "STRAT_O3", "STRAT_H2O"]

        df_temp = pd.read_csv(os.path.join(folder, "temp_forc.txt"), delimiter=r"\s+")
        years = df_temp.Year[:]
        if variable in forc_sums:
            timeseries = np.zeros(len(years))
            for comp, value in self.variable_dict.items():
                if variable in comp and value not in forc_sums:
                    timeseries = timeseries + df_temp[value].to_numpy()
        elif variable in ("Fgas", "GHG"):
            timeseries = np.zeros(len(years))
            for comp in fgas_list:
                timeseries = timeseries + df_temp[comp].to_numpy()
            if variable == "GHG":
                for comp in ghg_not_fgas:
                    timeseries = timeseries + df_temp[comp].to_numpy()
        elif variable == "Total_forcing+sunvolc":
            timeseries = df_temp["Total_forcing"].to_numpy()
            timeseries = timeseries + self.get_volc_forcing(
                years.to_numpy()[0], years.to_numpy()[-1]
            )
            timeseries = timeseries + self.get_sun_forcing(
                years.to_numpy()[0], years.to_numpy()[-1]
            )
        else:
            timeseries = df_temp[variable].to_numpy()
        return years, timeseries



import os
from unittest import mock

import pytest

from openscm_runner.settings import ConfigLoader


@mock.patch("openscm_runner.settings.dotenv_values")
def test_config_empty(mock_dotenv, monkeypatch):
    mock_dotenv.return_value = {}
    # This assumes that the envvar TEST_ENV_VAR does not exist
    config = ConfigLoader()
    assert not config.is_loaded

    with pytest.raises(KeyError, match="TEST_ENV_VAR"):
        config["TEST_ENV_VAR"]

    with pytest.raises(KeyError, match="TEST_ENV_VAR"):
        config["test_ENV_var"]


@mock.patch("openscm_runner.settings.dotenv_values")
def test_config_envvar(mock_dotenv, monkeypatch):
    mock_dotenv.return_value = {}
    config = ConfigLoader()

    res = config["PATH"]
    assert res == os.environ["PATH"]

    res = config["paTh"]
    assert res == os.environ["PATH"]


@mock.patch("openscm_runner.settings.dotenv_values")
def test_config_with_dotenv(mock_dotenv, monkeypatch):
    monkeypatch.setenv("TEST", "env_var")
    mock_dotenv.return_value = {"TEST": "env_file", "EXTRA": "yay"}

    config = ConfigLoader()

    assert config["TEST"] == "env_var"
    assert config["EXTRA"] == "yay"
    assert config["extra"] == "yay"

    with pytest.raises(KeyError, match="NOT_THERE"):
        config["not_there"]

    assert config.is_loaded
    mock_dotenv.assert_called_once()
    assert config._config == mock_dotenv.return_value


@mock.patch("openscm_runner.settings.dotenv_values")
def test_config_get(mock_dotenv, monkeypatch):
    monkeypatch.setenv("TEST", "env_var")
    mock_dotenv.return_value = {}

    config = ConfigLoader()

    assert config.get("TEST", "default") == "env_var"
    assert config.get("UNKNOWN_CONFIG", "default") == "default"



"""
CICEROSCM_WRAPPER for parallelisation
"""
import logging
import os
import re
import shutil
import subprocess  # nosec # have to use subprocess
import tempfile
from distutils import dir_util

import pandas as pd
from scmdata import ScmRun, run_append

from ...settings import config
from ._utils import _get_unique_index_values
from .make_scenario_files import SCENARIOFILEWRITER
from .read_results import CSCMREADER
from .write_parameter_files import PARAMETERFILEWRITER

LOGGER = logging.getLogger(__name__)


class CiceroSCMWrapper:  # pylint: disable=too-few-public-methods
    """
    CICEROSCM Wrapper for parallel runs
    """

    def __init__(self, scenariodata):
        """
        Intialise CICEROSCM wrapper
        """
        self.udir = os.path.join(os.path.dirname(__file__), "utils_templates")
        self.sfilewriter = SCENARIOFILEWRITER(self.udir)
        self.pamfilewriter = PARAMETERFILEWRITER(self.udir)
        self._setup_tempdirs()
        self.resultsreader = CSCMREADER(self.rundir)

        self.scen = _get_unique_index_values(scenariodata, "scenario")
        self.model = _get_unique_index_values(scenariodata, "model")
        self._make_dir_structure(re.sub("[^a-zA-Z0-9_-]", "", self.scen)[:50])
        self._call_sfilewriter(scenariodata)

    def _call_sfilewriter(self, scenarios):
        """
        Call sfilwriter to write scenariodata file
        """
        self.sfilewriter.write_scenario_data(
            scenarios,
            os.path.join(self.rundir, re.sub("[^a-zA-Z0-9_-]", "", self.scen)[:50]),
        )

    def run_over_cfgs(self, cfgs, output_variables):
        """
        Run over each configuration parameter set
        write parameterfiles, run, read results
        and make an ScmRun with results
        """
        runs = []
        for i, pamset in enumerate(cfgs):
            self.pamfilewriter.write_parameterfile(
                pamset,
                os.path.join(self.rundir, re.sub("[^a-zA-Z0-9_-]", "", self.scen)[:50]),
            )
            executable = os.path.join(self.rundir, "scm_vCH4fb")
            pamfile = os.path.join(
                self.rundir,
                re.sub("[^a-zA-Z0-9_-]", "", self.scen)[:50],
                "inputfiles",
                "pam_current.scm",
            )
            call = f"{executable} {pamfile}"

            LOGGER.debug("Call, %s", call)
            subprocess.check_call(
                call, cwd=self.rundir, shell=True,  # nosec # have to use subprocess
            )
            for variable in output_variables:
                (
                    years,
                    timeseries,
                    unit,
                ) = self.resultsreader.read_variable_timeseries(
                    self.scen, variable, self.sfilewriter
                )
                if years.empty:  # pragma: no cover
                    continue  # pragma: no cover

                runs.append(
                    ScmRun(
                        pd.Series(timeseries, index=years),
                        columns={
                            "climate_model": "CICERO-SCM",
                            "model": self.model,
                            "run_id": pamset.get("Index", i),
                            "scenario": self.scen,
                            "region": ["World"],
                            "variable": [variable],
                            "unit": [unit],
                        },
                    )
                )

        return run_append(runs)

    def _setup_tempdirs(self):
        """
        Set up temporary directories to run and make output in
        """
        root_dir = config.get("CICEROSCM_WORKER_ROOT_DIR", None)
        self.rundir = tempfile.mkdtemp(prefix="ciceroscm-", dir=root_dir)
        LOGGER.info("Creating new CICERO-SCM instance: %s", self.rundir)
        dir_util.copy_tree(
            os.path.join(os.path.dirname(__file__), "utils_templates", "run_dir"),
            self.rundir,
        )

    def cleanup_tempdirs(self):
        """
        Remove tempdirs after run
        """
        LOGGER.info("Removing CICERO-SCM instance: %s", self.rundir)
        shutil.rmtree(self.rundir)

    def _make_dir_structure(self, scenario):
        """
        Make directory structure for a scenario in which to put input and
        outputfiles for the run
        """
        os.makedirs(self.rundir, exist_ok=True)
        os.makedirs(os.path.join(self.rundir, scenario), exist_ok=True)
        os.makedirs(os.path.join(self.rundir, scenario, "inputfiles"), exist_ok=True)
        os.makedirs(os.path.join(self.rundir, scenario, "outputfiles"), exist_ok=True)



"""
OpenSCM-Runner, a thin wrapper to run simple climate models with a unified interface.

See README and docs for more info.
"""
from .run import run  # noqa: F401

try:
    from importlib.metadata import version as _version
except ImportError:
    # no recourse if the fallback isn't there either...
    from importlib_metadata import version as _version

try:
    __version__ = _version("openscm_runner")
except Exception:  # pylint: disable=broad-except
    # Local copy, not installed with setuptools
    __version__ = "unknown"



"""
Utility functions
"""
import scmdata


def calculate_quantiles(
    scmdf,
    quantiles,
    process_over_columns=("run_id", "ensemble_member", "climate_model"),
):
    """
    Calculate quantiles of an :obj:`ScmRun`

    Parameters
    ----------
    scmdf : :obj:`ScmRun`
        :obj:`ScmRun` containing the data from which to calculate the
        quantiles

    quantiles : list of float
        quantiles to calculate (must be in [0, 1])

    process_over_columns : list of str
        Columns to process over. All other columns in ``scmdf.meta`` will be
        included in the output.

    Returns
    -------
    :obj:`ScmRun`
        :obj:`ScmRun` containing the quantiles of interest, processed
        over ``process_over_columns``
    """
    out = []
    for quant in quantiles:
        quantile_df = scmdf.process_over(process_over_columns, "quantile", q=quant)
        quantile_df["quantile"] = quant

        out.append(quantile_df)

    out = scmdata.run_append([scmdata.ScmRun(o) for o in out])

    return out



"""
Handling of compatibility of MAGICC imports with different states of installation
"""
# pylint:disable=unused-import
try:
    import f90nml
    import pymagicc

    HAS_PYMAGICC = True
except ImportError:
    f90nml = None
    pymagicc = None
    HAS_PYMAGICC = False



"""
Rather than hard-coding constants, configuration can be source from 2 different
sources:

* Environment variables
* dotenv files

Environment variables are always used in preference to the values in dotenv
files.
"""


import os

from dotenv import dotenv_values, find_dotenv


class ConfigLoader:
    """
    Configuration container

    Loads a local dotenv file containing configuration. An example of a
    configuration file is provided in root of the project.

    A combination of environment variables and dotenv files can be used as
    configuration with an existing environment variables overriding a value
    specified in a dotenv file.

    All configuration is case-insensitive.

    .. code:: python

        >>> config = ConfigLoader()
        >>> config['VALUE']
        >>> config.get("VALUE", None)
    """

    def __init__(self):
        self._config = {}
        self.is_loaded = False

    def load_config(self):
        """
        Load configuration from disk

        A dotenv file is attempted to be loaded. The first file named .env
        in the current directory or any of its parents will read in.

        If no dotenv files are found, then
        """
        dotenv_cfg = dotenv_values(find_dotenv(usecwd=True), verbose=True)
        self.update(dotenv_cfg)

        # Add any extra files here

        self.is_loaded = True

    def get(self, key, default=None):
        """
        Get value for a given key, falling back to a default value if not present.

        Parameters
        ----------
        key : str
            Key

        default: Any
            Default value returned if no configuration for ``key`` is present.

        Returns
        -------
        Any
            Value with key ``item``. If not value is present ``default``
            is returned.
        """
        try:
            return self[key]
        except KeyError:
            return default

    def __getitem__(self, key):
        """
        Get a config value for a given key

        Parameters
        ----------
        key: str
            Key

        Returns
        -------
        Any
            Value with key ``item``

        Raises
        ------
        KeyError
            No configuration values for the key were found
        """
        if not self.is_loaded:
            # Lazy loaded
            self.load_config()
        key = key.upper()

        # Always preference environment variable
        if key in os.environ:
            return os.environ[key]

        # Fall back to loaded config
        # Preference determined by load order
        if key in self._config:
            return self._config[key]

        # Raise KeyError if not found
        raise KeyError(key)

    def update(self, conf):
        """
        Update the configuration

        If configuration with duplicate keys already exists, then these values
        will overwrite the existing values.

        Parameters
        ----------
        conf: dict
            Configuration to use to update the config
        """
        conf = {k.upper(): v for k, v in conf.items()}
        self._config.update(conf)


config = ConfigLoader()



"""
Module with functionality to make emission input files
"""

# TODO: optimise to speed up reading and writing

import csv
import logging
import os
import re

import numpy as np
import openscm_units
import pandas as pd

from ._utils import _get_unique_index_values

LOGGER = logging.getLogger(__name__)


def _read_ssp245_em(ssp245_em_file):
    """
    Get default data from ssp245_RCMIP
    """
    ssp245df = (
        pd.read_csv(ssp245_em_file, delimiter="\t", index_col=0)
        .rename(columns=lambda x: x.strip())
        .rename(index=lambda x: x.strip())
        .rename(columns={"CO2 .1": "CO2_lu"})
    )

    return ssp245df


def _unit_conv_factor(unit, cicero_unit):
    with openscm_units.unit_registry.context("NOx_conversions"):

        if cicero_unit.startswith("GgH1"):
            conv_factor = (
                openscm_units.unit_registry(unit)
                .to(cicero_unit.replace("GgH1", "GgHalon1"))
                .magnitude
            )
        elif cicero_unit.startswith("GgH2"):
            conv_factor = (
                openscm_units.unit_registry(unit)
                .to(cicero_unit.replace("GgH2", "GgHalon2"))
                .magnitude
            )
        else:
            conv_factor = openscm_units.unit_registry(unit).to(cicero_unit).magnitude

    return conv_factor


class SCENARIOFILEWRITER:
    """
    Class to write scenariofiles:
    """

    def __init__(self, udir):
        self.components = []
        self.units = []
        self.concunits = []

        self.component_dict = {
            "CO2_lu": ["CO2|MAGICC AFOLU", 1],
            "CFC-113": ["CFC113", 1],
            "CFC-114": ["CFC114", 1],
            "SO2": ["Sulfur", 1],
            "NMVOC": ["VOC", 1],
            "CFC-11": ["CFC11", 1],
            "CFC-115": ["CFC115", 1],
            "CFC-12": ["CFC12", 1],
            "HCFC-141b": ["HCFC141b", 1],
            "HCFC-142b": ["HCFC142b", 1],
            "HCFC-22": ["HCFC22", 1],
            "H-1211": ["Halon1211", 1],
            "H-1301": ["Halon1301", 1],
            "H-2402": ["Halon2402", 1],
            "CO2": ["CO2|MAGICC Fossil and Industrial", 1],
            "CH4": ["CH4", 1],
            "N2O": ["N2O", 1],
            "CH3Br": ["CH3Br", 1],
            "CCl4": ["CCl4", 1],
            "CH3CCl3": ["CH3CCl3", 1],
            "HCFC-123": ["HCFC-123", 1],
            "HFC125": ["HFC125", 1],
            "HFC134a": ["HFC134a", 1],
            "HFC143a": ["HFC143a", 1],
            "HFC227ea": ["HFC227ea", 1],
            "HFC23": ["HFC23", 1],
            "HFC245fa": ["HFC245fa", 1],
            "HFC32": ["HFC32", 1],
            "HFC4310mee": ["HFC4310mee", 1],
            "C2F6": ["C2F6", 1],
            "C6F14": ["C6F14", 1],
            "CF4": ["CF4", 1],
            "SF6": ["SF6", 1],
            "NOx": ["NOx", 1],
            "CO": ["CO", 1],
            "NH3": ["NH3", 1],
            "BMB_AEROS_BC": ["BMB_AEROS_BC", 1],
            "BMB_AEROS_OC": ["BMB_AEROS_OC", 1],
            "BC": ["BC", 1],
            "OC": ["OC", 1],
        }  # Halon1212, CH3Cl
        self.initialize_units_comps(os.path.join(udir, "gases_v1RCMIP.txt"))
        self.years = np.arange(2015, 2101)  # Temporary default values, is updated later
        self.ssp245data = _read_ssp245_em(os.path.join(udir, "ssp245_em_RCMIP.txt"))
        self.udir = udir

    def get_top_of_file(self, ssp245_em_file):
        """
        Get the top of the emission file which will be equal for all scenarios
        """
        with open(ssp245_em_file, encoding="ascii") as semfile:
            filedata = semfile.read()
            top_of_file = filedata.split(f"\n{self.years[0]}")[0]

        return top_of_file

    def initialize_units_comps(self, gasfile):
        """
        Get the list of gas components and units
        from the gases file:
        """
        with open(gasfile, "r", encoding="ascii") as txt_rcpfile:
            gasreader = csv.reader(txt_rcpfile, delimiter="\t")
            next(gasreader)
            for row in gasreader:
                if row[1] == "X":
                    continue

                component = row[0]
                unit = row[1]

                if component == "N2O" and unit == "Tg_N":
                    # in openscm-units, to get the mass of nitrogen, have
                    # to use the unit "Tg N2ON" (converting to "Tg N" just
                    # converts using the mass fraction of a single nitrogen
                    # atom, admittedly this isn't immediately obvious and
                    # arguably is a bug in openscm-units)
                    unit = "TgN2ON"
                elif "_" in unit:
                    unit = unit.replace("_", "")
                else:
                    comp_str = component.replace("-", "").replace("BMB_AEROS_", "")
                    unit = f"{unit}{comp_str}"

                unit = f"{unit} / yr"

                self.components.append(component)
                self.units.append(unit)
                self.concunits.append(row[2])

        self.components.insert(1, "CO2_lu")
        self.units.insert(1, "PgC / yr")
        self.concunits.insert(1, "ppm")

    def get_unit_convfactor(self, comp, scenarioframe):
        """
        Get unit and conversion factor for component
        """
        # Find the unit and the original unit
        cicero_unit = self.units[self.components.index(comp)]
        unit = _get_unique_index_values(
            scenarioframe[
                scenarioframe.index.get_level_values("variable")
                == f"Emissions|{self.component_dict[comp][0]}"
            ],
            "unit",
        )

        return _unit_conv_factor(unit, cicero_unit)

    def transform_scenarioframe(self, scenarioframe):
        """
        Get rid of multiindex and interpolate scenarioframe
        """
        if (
            _get_unique_index_values(scenarioframe, "region") != "World"
        ):  # pragma: no cover
            raise NotImplementedError()  # emergency valve

        scenarioframe = scenarioframe.reset_index(
            ("model", "region", "scenario", "unit"), drop=True
        )
        years = scenarioframe.columns

        if not isinstance(years[0], np.int64):
            yearsint = [np.int64(d.year) for d in years]
            scenarioframe.rename(
                lambda d: np.int64(d.year), axis="columns", inplace=True
            )
        else:
            yearsint = years

        self.years = np.arange(yearsint[0], yearsint[-1] + 1)
        for year in self.years:
            if year not in scenarioframe.columns:
                scenarioframe[year] = np.nan

        scenarioframe = scenarioframe.reindex(sorted(scenarioframe.columns), axis=1)
        interpol = scenarioframe.interpolate(axis=1)

        return interpol

    def write_scenario_data(self, scenarioframe, odir):
        """
        Take a scenariodataframe
        and writing out necessary emissions files
        """
        scenario = re.sub(
            "[^a-zA-Z0-9_-]", "", _get_unique_index_values(scenarioframe, "scenario"),
        )[:50]
        fname = os.path.join(odir, "inputfiles", f"{scenario}_em.txt",)
        logging.getLogger("pyam").setLevel(logging.ERROR)
        avail_comps = [
            c.replace("Emissions|", "")
            for c in _get_unique_index_values(
                scenarioframe, "variable", assert_all_same=False
            )
        ]
        ciceroscm_comps = [v[0] for v in self.component_dict.values()]
        not_used_comps = set(avail_comps) - set(ciceroscm_comps)
        if not_used_comps:
            LOGGER.warning("%s not used by CICERO-SCM", not_used_comps)

        interpol = self.transform_scenarioframe(scenarioframe)
        printout_frame = pd.DataFrame(columns=self.components, index=self.years)

        # Setting conversion factors for components with data from scenarioframe
        for comp in self.components:
            if self.component_dict[comp][0] in avail_comps:
                convfactor = self.get_unit_convfactor(comp, scenarioframe)
                if (
                    self.component_dict[comp][0] in ("BC", "OC")
                    and f"BMB_AEROS_{self.component_dict[comp][0]}" not in avail_comps
                ):
                    printout_frame[comp] = (
                        interpol.T[f"Emissions|{self.component_dict[comp][0]}"]
                        * convfactor
                    ).to_numpy() - self.ssp245data[
                        f"BMB_AEROS_{self.component_dict[comp][0]}"
                    ].loc[
                        str(self.years[0]) : str(self.years[-1])
                    ].to_numpy().astype(
                        np.float
                    )
                else:
                    printout_frame[comp] = (
                        interpol.T[f"Emissions|{self.component_dict[comp][0]}"]
                        * convfactor
                    )
            else:
                LOGGER.warning("No %s data available, using ssp245", comp)
                printout_frame[comp] = (
                    self.ssp245data[comp]
                    .loc[str(self.years[0]) : str(self.years[-1])]
                    .to_numpy()
                )

        printout_frame = printout_frame.astype(float).reset_index()
        printout_frame_fmt = ["%d"] + ["%.8f"] * (printout_frame.shape[1] - 1)

        with open(fname, "w", encoding="ascii") as sfile:
            sfile.write(
                self.get_top_of_file(os.path.join(self.udir, "ssp245_em_RCMIP.txt"))
            )
            sfile.write("\n")
            np.savetxt(sfile, printout_frame, fmt=printout_frame_fmt, delimiter=" \t ")



"""
Handling of compatibility of fair imports with different states of installation
"""
# pylint:disable=unused-import
try:
    import fair
    from fair.forward import fair_scm

    HAS_FAIR = True
except ImportError:
    fair = None
    fair_scm = None
    HAS_FAIR = False



"""
Base class for adapters
"""
from abc import ABC, abstractmethod


class _Adapter(ABC):  # pylint: disable=too-few-public-methods
    """
    Base class for adapters
    """

    def __init__(self, *args, **kwargs):
        """
        Initialise the adapter

        Parameters
        ----------
        *args
            Passed to the adapter's ``_init_model`` method

        **kwargs
            Passed to the adapter's ``_init_model`` method
        """
        self._init_model(*args, **kwargs)

    @abstractmethod
    def _init_model(self, *args, **kwargs):
        """
        Initialise the model

        Parameters
        ----------
        *args
            Arguments used to initialise the model

        **kwargs
            Keyword arguments used to initialise the model
        """

    def run(self, scenarios, cfgs, output_variables, output_config):
        """
        Parameters
        ----------
        scenarios : :obj:`pyam.IamDataFrame`
            Scenarios to run

        cfgs : list[dict]
            The config with which to run the model

        output_variables : list[str]
            Variables to include in the output

        output_config : tuple[str]
            Configuration to include in the output

        Returns
        -------
        :obj:`ScmRun`
            Model output
        """
        return self._run(scenarios, cfgs, output_variables, output_config)

    @abstractmethod
    def _run(self, scenarios, cfgs, output_variables, output_config):
        """
        Run the model.

        This method is the internal implementation of the :meth:`run` interface
        """



"""
Module for running MAGICC in parallel
"""
import logging
import multiprocessing
import os.path
from concurrent.futures import ProcessPoolExecutor
from subprocess import CalledProcessError  # nosec

import scmdata

from ...settings import config
from ..utils._parallel_process import _parallel_process
from ._compat import f90nml, pymagicc
from ._magicc_instances import _MagiccInstances

LOGGER = logging.getLogger(__name__)


def _inject_pymagicc_compatible_magcfg_user(magicc):
    """
    Overwrite ``magicc.run_dir / MAGCFG_USER.CFG`` with config that only point to ``MAGTUNE_PYMAGICC.CFG``

    Parameters
    ----------
    magicc : :obj:`pymagicc.MAGICC7`
        Instance of :obj:`pymagicc.MAGICC7` to setup
    """
    LOGGER.info("Writing Pymagicc compatible MAGCFG_USER.CFG in %s", magicc.run_dir)
    with open(
        os.path.join(magicc.run_dir, "MAGCFG_USER.CFG"), "w", encoding="ascii"
    ) as file_handle:
        f90nml.write({"nml_allcfgs": {"file_tuningmodel_1": "PYMAGICC"}}, file_handle)


def _setup_func(magicc):
    LOGGER.info(
        "Setting up MAGICC worker in %s", magicc.root_dir,
    )

    magicc.set_config()

    _inject_pymagicc_compatible_magcfg_user(magicc)


def _init_magicc_worker(dict_shared_instances):
    LOGGER.debug("Initialising process %s", multiprocessing.current_process())
    LOGGER.debug("Existing instances %s", dict_shared_instances)
    magicc_instances = _MagiccInstances(  # noqa: F841 # pylint:disable=unused-variable
        dict_shared_instances
    )


def _run_func(magicc, cfg):
    try:
        scenario = cfg.pop("scenario")
        model = cfg.pop("model")
        output_config = cfg.pop("output_config")

        res = magicc.run(**cfg)
        if res.metadata["stderr"]:
            LOGGER.info("magicc run stderr: %s", res.metadata["stderr"])
            LOGGER.info("cfg: %s", cfg)

        res["scenario"] = scenario
        res["model"] = model
        res["run_id"] = cfg["run_id"]
        if output_config is not None:
            magicc_out_cfg = res.metadata["parameters"]["allcfgs"]
            for k in output_config:
                res[k] = cfg[k]
                if k in magicc_out_cfg:
                    if magicc_out_cfg[k] != cfg[k]:
                        LOGGER.warning(
                            "Parameter: %s. "
                            "MAGICC input config (via OpenSCM-Runner): %s. "
                            "MAGICC output config: %s.",
                            k,
                            cfg[k],
                            magicc_out_cfg[k],
                        )

        return res
    except CalledProcessError as exc:
        # Swallow the exception, but return None
        LOGGER.debug("magicc run failed: %s", exc.stderr)
        LOGGER.debug("cfg: %s", cfg)

        return None


def _execute_run(cfg, run_func, setup_func, instances):
    magicc = instances.get(
        root_dir=config["MAGICC_WORKER_ROOT_DIR"],
        init_callback=setup_func,
        init_callback_kwargs={},
    )

    return run_func(magicc, cfg)


def run_magicc_parallel(cfgs, output_vars, output_config):
    """
    Run MAGICC in parallel using compact out files

    Parameters
    ----------
    cfgs : list[dict]
        List of configurations with which to run MAGICC

    output_vars : list[str]
        Variables to output

    output_config : tuple[str]
        Configuration to include in the output

    Returns
    -------
    :obj:`ScmRun`
        :obj:`ScmRun` instance with all results.
    """
    LOGGER.info("Entered _parallel_magicc_compact_out")
    shared_manager = multiprocessing.Manager()
    shared_dict = shared_manager.dict()
    instances = _MagiccInstances(existing_instances=shared_dict)

    magicc_internal_vars = [
        f"DAT_{pymagicc.definitions.convert_magicc7_to_openscm_variables(v, inverse=True)}"
        for v in output_vars
    ]

    runs = [
        {
            "cfg": {
                **cfg,
                "only": output_vars,
                "out_dynamic_vars": magicc_internal_vars,
                "output_config": output_config,
            },
            "run_func": _run_func,
            "setup_func": _setup_func,
            "instances": instances,
        }
        for cfg in cfgs
    ]

    try:
        max_workers = int(
            config.get("MAGICC_WORKER_NUMBER", multiprocessing.cpu_count())
        )
        LOGGER.info("Running in parallel with up to %d workers", max_workers)
        pool = ProcessPoolExecutor(  # need to handle shared_manager too
            max_workers=max_workers,
            initializer=_init_magicc_worker,
            initargs=(shared_dict,),
        )

        res = _parallel_process(
            func=_execute_run,
            configuration=runs,
            pool=pool,
            config_are_kwargs=True,
            front_serial=2,
            front_parallel=2,
        )

        LOGGER.info("Appending results into a single ScmRun")
        res = scmdata.run_append([r for r in res if r is not None])

    finally:
        instances.cleanup()
        LOGGER.info("Shutting down parallel pool")
        shared_manager.shutdown()
        pool.shutdown()

    return res



import os.path
import requests

import pyam
import pymagicc.definitions

RCMIP_EMISSIONS_URL = "https://drive.google.com/u/0/uc?id=1krA0lficstXqahlNCko7nbfqgjKrd_Sa&export=download"
DATA_FILE_RAW_NAME = "rcmip_emissions.csv"
DATA_FILE_OUT_NAME = "rcmip_scen_ssp_world_emissions.csv"

data_file_raw = os.path.join(os.path.dirname(__file__), DATA_FILE_RAW_NAME)
data_file_out = os.path.join(os.path.dirname(__file__), DATA_FILE_OUT_NAME)

if not os.path.isfile(data_file_raw):
    request = requests.get(RCMIP_EMISSIONS_URL, allow_redirects=True)
    with open(data_file_raw, "wb") as fh:
        fh.write(request.content)

keep_vars = pymagicc.definitions.convert_magicc7_to_openscm_variables(
    [
        "{}_EMIS".format(v)
        for v in pymagicc.definitions.PART_OF_SCENFILE_WITH_EMISSIONS_CODE_1
    ]
)
keep_vars = [
    v.replace("HFC4310", "HFC4310mee").replace("SOx", "Sulfur").replace("NMVOC", "VOC")
    for v in keep_vars
]
data = pyam.IamDataFrame(data_file_raw)
data_out = data.filter(
    scenario="ssp*", region="World", year=range(2015, 2101)
).timeseries()
data_out.index = data_out.index.droplevel(["activity_id", "mip_era"])
data_out = data_out.reset_index()
data_out["variable"] = data_out["variable"].apply(
    lambda x: x.replace("Montreal Gases|", "")
    .replace("F-Gases|HFC|", "")
    .replace("F-Gases|PFC|", "")
    .replace("F-Gases|CFC|", "")
    .replace("F-Gases|", "")
    .replace("CFC|", "")
)

data_out = pyam.IamDataFrame(data_out).filter(variable=keep_vars)
data_out.to_csv(data_file_out)



"""
Module for running FaIR
"""
import logging
import multiprocessing
from concurrent.futures import ProcessPoolExecutor

import numpy as np
from scmdata import ScmRun, run_append

from ...settings import config
from ..utils._parallel_process import _parallel_process
from ._compat import fair_scm

LOGGER = logging.getLogger(__name__)


def run_fair(cfgs, output_vars):  # pylint: disable=R0914
    """
    Run FaIR

    Parameters
    ----------
    cfgs : list[dict]
        List of configurations with which to run FaIR

    output_vars : list[str]
        Variables to output

    Returns
    -------
    :obj:`ScmRun`
        :obj:`ScmRun` instance with all results.
    """
    res = []
    updated_config = []
    for i, cfg in enumerate(cfgs):
        updated_config.append({})
        for key, value in cfg.items():
            if isinstance(value, list):
                updated_config[i][key] = np.asarray(value)
            else:
                updated_config[i][key] = value
        updated_config[i]["output_vars"] = output_vars

    ncpu = int(config.get("FAIR_WORKER_NUMBER", multiprocessing.cpu_count()))
    LOGGER.info("Running FaIR with %s workers", ncpu)

    parallel_process_kwargs = dict(
        func=_single_fair_iteration,
        configuration=updated_config,
        config_are_kwargs=False,
    )
    if ncpu > 1:

        with ProcessPoolExecutor(ncpu) as pool:
            res = _parallel_process(**parallel_process_kwargs, pool=pool,)
    else:
        res = _parallel_process(**parallel_process_kwargs)

    res = run_append(res)

    return res


def _single_fair_iteration(cfg):  # pylint: disable=R0914
    scenario = cfg.pop("scenario")
    model = cfg.pop("model")
    run_id = cfg.pop("run_id")
    factors = {}
    factors["gmst"] = cfg.pop("gmst_factor")
    factors["ohu"] = cfg.pop("ohu_factor")
    startyear = cfg.pop("startyear")
    output_vars = cfg.pop("output_vars")

    data, unit, nt = _process_output(fair_scm(**cfg), output_vars, factors)

    data_scmrun = []
    variables = []
    units = []
    for key, variable in data.items():
        variables.append(key)
        data_scmrun.append(variable)
        units.append(unit[key])

    tempres = ScmRun(
        np.vstack(data_scmrun).T,
        index=np.arange(startyear, startyear + nt),
        columns={
            "scenario": scenario,
            "model": model,
            "region": "World",
            "variable": variables,
            "unit": units,
            "run_id": run_id,
        },
    )

    return tempres


def _process_output(fair_output, output_vars, factors):  # pylint: disable=R0915
    """
    Make sense of FaIR1.6 output

    Parameters
    ----------
    fair_output : tuple
        7-tuple of concentrations, forcing, temperature, lambda_eff, ohc, heatflux, airborne_emissions:
            c : np.ndarray
                (nt, 31) array of greenhouse gas concentrations
            f : np.ndarray
                (nt, 41) array of effective radiative forcings
            t : np.ndarray
                (nt,) array of temperature
            lambda_eff: np.ndarray
                effective climate feedback
            ohc : np.ndarray
                total ocean heat uptake
            heatflux:
                heat transfer into the ocean
            airborne_emissions:
                atmospheric carbon content

    output_vars : list[str]
        List of output variables

    factors : dict[(Union[float, numpy.ndarray])]
        ohu : ratio of ocean heat uptake to total Earth energy uptake
        gmst : ratio of GMST to GSAT

    Returns
    -------
    data : dict
        dict of climate model output
    unit : dict
        dict of units corresponding to data
    nt : int
        number of timesteps modelled
    """
    (
        concentrations,
        forcing,
        temperature,
        lambda_eff,
        ohc,
        heatflux,
        airborne_emissions,
    ) = fair_output

    data = {}
    unit = {}

    data["Atmospheric Concentrations|CO2"] = concentrations[:, 0]
    data["Atmospheric Concentrations|CH4"] = concentrations[:, 1]
    data["Atmospheric Concentrations|N2O"] = concentrations[:, 2]
    data["Atmospheric Concentrations|CF4"] = concentrations[:, 3]
    data["Atmospheric Concentrations|C2F6"] = concentrations[:, 4]
    data["Atmospheric Concentrations|C6F14"] = concentrations[:, 5]
    data["Atmospheric Concentrations|HFC23"] = concentrations[:, 6]
    data["Atmospheric Concentrations|HFC32"] = concentrations[:, 7]
    data["Atmospheric Concentrations|HFC125"] = concentrations[:, 8]
    data["Atmospheric Concentrations|HFC134a"] = concentrations[:, 9]
    data["Atmospheric Concentrations|HFC143a"] = concentrations[:, 10]
    data["Atmospheric Concentrations|HFC227ea"] = concentrations[:, 11]
    data["Atmospheric Concentrations|HFC245fa"] = concentrations[:, 12]
    data["Atmospheric Concentrations|HFC4310mee"] = concentrations[:, 13]
    data["Atmospheric Concentrations|SF6"] = concentrations[:, 14]
    data["Atmospheric Concentrations|CFC11"] = concentrations[:, 15]
    data["Atmospheric Concentrations|CFC12"] = concentrations[:, 16]
    data["Atmospheric Concentrations|CFC113"] = concentrations[:, 17]
    data["Atmospheric Concentrations|CFC114"] = concentrations[:, 18]
    data["Atmospheric Concentrations|CFC115"] = concentrations[:, 19]
    data["Atmospheric Concentrations|CCl4"] = concentrations[:, 20]
    data["Atmospheric Concentrations|CH3CCl3"] = concentrations[:, 21]
    data["Atmospheric Concentrations|HCFC22"] = concentrations[:, 22]
    data["Atmospheric Concentrations|HCFC141b"] = concentrations[:, 23]
    data["Atmospheric Concentrations|HCFC142b"] = concentrations[:, 24]
    data["Atmospheric Concentrations|Halon1211"] = concentrations[:, 25]
    data["Atmospheric Concentrations|Halon1202"] = concentrations[:, 26]
    data["Atmospheric Concentrations|Halon1301"] = concentrations[:, 27]
    data["Atmospheric Concentrations|Halon2402"] = concentrations[:, 28]
    data["Atmospheric Concentrations|CH3Br"] = concentrations[:, 29]
    data["Atmospheric Concentrations|CH3Cl"] = concentrations[:, 30]
    data["Effective Radiative Forcing|CO2"] = forcing[:, 0]
    data["Effective Radiative Forcing|CH4"] = forcing[:, 1]
    data["Effective Radiative Forcing|N2O"] = forcing[:, 2]
    data["Effective Radiative Forcing|CF4"] = forcing[:, 3]
    data["Effective Radiative Forcing|C2F6"] = forcing[:, 4]
    data["Effective Radiative Forcing|C6F14"] = forcing[:, 5]
    data["Effective Radiative Forcing|HFC23"] = forcing[:, 6]
    data["Effective Radiative Forcing|HFC32"] = forcing[:, 7]
    data["Effective Radiative Forcing|HFC125"] = forcing[:, 8]
    data["Effective Radiative Forcing|HFC134a"] = forcing[:, 9]
    data["Effective Radiative Forcing|HFC143a"] = forcing[:, 10]
    data["Effective Radiative Forcing|HFC227ea"] = forcing[:, 11]
    data["Effective Radiative Forcing|HFC245fa"] = forcing[:, 12]
    data["Effective Radiative Forcing|HFC4310mee"] = forcing[:, 13]
    data["Effective Radiative Forcing|SF6"] = forcing[:, 14]
    data["Effective Radiative Forcing|CFC11"] = forcing[:, 15]
    data["Effective Radiative Forcing|CFC12"] = forcing[:, 16]
    data["Effective Radiative Forcing|CFC113"] = forcing[:, 17]
    data["Effective Radiative Forcing|CFC114"] = forcing[:, 18]
    data["Effective Radiative Forcing|CFC115"] = forcing[:, 19]
    data["Effective Radiative Forcing|CCl4"] = forcing[:, 20]
    data["Effective Radiative Forcing|CH3CCl3"] = forcing[:, 21]
    data["Effective Radiative Forcing|HCFC22"] = forcing[:, 22]
    data["Effective Radiative Forcing|HCFC141b"] = forcing[:, 23]
    data["Effective Radiative Forcing|HCFC142b"] = forcing[:, 24]
    data["Effective Radiative Forcing|Halon1211"] = forcing[:, 25]
    data["Effective Radiative Forcing|Halon1202"] = forcing[:, 26]
    data["Effective Radiative Forcing|Halon1301"] = forcing[:, 27]
    data["Effective Radiative Forcing|Halon2402"] = forcing[:, 28]
    data["Effective Radiative Forcing|CH3Br"] = forcing[:, 29]
    data["Effective Radiative Forcing|CH3Cl"] = forcing[:, 30]
    data["Effective Radiative Forcing|Tropospheric Ozone"] = forcing[:, 31]
    data["Effective Radiative Forcing|Stratospheric Ozone"] = forcing[:, 32]
    data["Effective Radiative Forcing|CH4 Oxidation Stratospheric H2O"] = forcing[:, 33]
    data["Effective Radiative Forcing|Contrails"] = forcing[:, 34]
    data["Effective Radiative Forcing|Aerosols|Direct Effect|SOx"] = forcing[:, 35]
    data[
        "Effective Radiative Forcing|Aerosols|Direct Effect|Secondary Organic Aerosol"
    ] = forcing[:, 36]
    data["Effective Radiative Forcing|Aerosols|Direct Effect|Nitrate"] = forcing[:, 37]
    data["Effective Radiative Forcing|Aerosols|Direct Effect|BC"] = forcing[:, 38]
    data["Effective Radiative Forcing|Aerosols|Direct Effect|OC"] = forcing[:, 39]
    data["Effective Radiative Forcing|Aerosols|Indirect Effect"] = forcing[:, 40]
    data["Effective Radiative Forcing|Black Carbon on Snow"] = forcing[:, 41]
    data["Effective Radiative Forcing|Land-use Change"] = forcing[:, 42]
    data["Effective Radiative Forcing|Volcanic"] = forcing[:, 43]
    data["Effective Radiative Forcing|Solar"] = forcing[:, 44]
    data["Effective Radiative Forcing"] = np.sum(forcing, axis=1)
    data["Effective Radiative Forcing|Anthropogenic"] = np.sum(forcing[:, :43], axis=1)
    data["Effective Radiative Forcing|Greenhouse Gases"] = np.sum(
        forcing[:, :31], axis=1
    )
    # This definition does not include ozone and H2O from CH4 oxidation
    data["Effective Radiative Forcing|Kyoto Gases"] = np.sum(forcing[:, :15], axis=1)
    data["Effective Radiative Forcing|CO2, CH4 and N2O"] = np.sum(
        forcing[:, :3], axis=1
    )
    # What is the rigorous definition here? CFCs are not included but contain F
    data["Effective Radiative Forcing|F-Gases"] = np.sum(forcing[:, 3:15], axis=1)
    data["Effective Radiative Forcing|Montreal Protocol Halogen Gases"] = np.sum(
        forcing[:, 15:31], axis=1
    )
    data["Effective Radiative Forcing|Aerosols|Direct Effect"] = np.sum(
        forcing[:, 35:40], axis=1
    )
    data["Effective Radiative Forcing|Aerosols"] = np.sum(forcing[:, 35:41], axis=1)
    data["Effective Radiative Forcing|Ozone"] = np.sum(forcing[:, 31:33], axis=1)
    data["Surface Air Temperature Change"] = temperature
    data["Surface Air Ocean Blended Temperature Change"] = temperature * factors["gmst"]
    data["Airborne Fraction"] = airborne_emissions
    data["Effective Climate Feedback"] = lambda_eff
    data["Heat Content"] = ohc
    data["Heat Content|Ocean"] = ohc * factors["ohu"]
    data["Net Energy Imbalance"] = heatflux
    data["Heat Uptake"] = heatflux
    data["Heat Uptake|Ocean"] = heatflux * factors["ohu"]

    unit["Atmospheric Concentrations|CO2"] = "ppm"
    unit["Atmospheric Concentrations|CH4"] = "ppb"
    unit["Atmospheric Concentrations|N2O"] = "ppb"
    unit["Atmospheric Concentrations|CF4"] = "ppt"
    unit["Atmospheric Concentrations|C2F6"] = "ppt"
    unit["Atmospheric Concentrations|C6F14"] = "ppt"
    unit["Atmospheric Concentrations|HFC23"] = "ppt"
    unit["Atmospheric Concentrations|HFC32"] = "ppt"
    unit["Atmospheric Concentrations|HFC125"] = "ppt"
    unit["Atmospheric Concentrations|HFC134a"] = "ppt"
    unit["Atmospheric Concentrations|HFC143a"] = "ppt"
    unit["Atmospheric Concentrations|HFC227ea"] = "ppt"
    unit["Atmospheric Concentrations|HFC245fa"] = "ppt"
    unit["Atmospheric Concentrations|HFC4310mee"] = "ppt"
    unit["Atmospheric Concentrations|SF6"] = "ppt"
    unit["Atmospheric Concentrations|CFC11"] = "ppt"
    unit["Atmospheric Concentrations|CFC12"] = "ppt"
    unit["Atmospheric Concentrations|CFC113"] = "ppt"
    unit["Atmospheric Concentrations|CFC114"] = "ppt"
    unit["Atmospheric Concentrations|CFC115"] = "ppt"
    unit["Atmospheric Concentrations|CCl4"] = "ppt"
    unit["Atmospheric Concentrations|CH3CCl3"] = "ppt"
    unit["Atmospheric Concentrations|HCFC22"] = "ppt"
    unit["Atmospheric Concentrations|HCFC141b"] = "ppt"
    unit["Atmospheric Concentrations|HCFC142b"] = "ppt"
    unit["Atmospheric Concentrations|Halon1211"] = "ppt"
    unit["Atmospheric Concentrations|Halon1202"] = "ppt"
    unit["Atmospheric Concentrations|Halon1301"] = "ppt"
    unit["Atmospheric Concentrations|Halon2402"] = "ppt"
    unit["Atmospheric Concentrations|CH3Br"] = "ppt"
    unit["Atmospheric Concentrations|CH3Cl"] = "ppt"
    unit["Effective Radiative Forcing|CO2"] = "W/m**2"
    unit["Effective Radiative Forcing|CH4"] = "W/m**2"
    unit["Effective Radiative Forcing|N2O"] = "W/m**2"
    unit["Effective Radiative Forcing|CF4"] = "W/m**2"
    unit["Effective Radiative Forcing|C2F6"] = "W/m**2"
    unit["Effective Radiative Forcing|C6F14"] = "W/m**2"
    unit["Effective Radiative Forcing|HFC23"] = "W/m**2"
    unit["Effective Radiative Forcing|HFC32"] = "W/m**2"
    unit["Effective Radiative Forcing|HFC125"] = "W/m**2"
    unit["Effective Radiative Forcing|HFC134a"] = "W/m**2"
    unit["Effective Radiative Forcing|HFC143a"] = "W/m**2"
    unit["Effective Radiative Forcing|HFC227ea"] = "W/m**2"
    unit["Effective Radiative Forcing|HFC245fa"] = "W/m**2"
    unit["Effective Radiative Forcing|HFC4310mee"] = "W/m**2"
    unit["Effective Radiative Forcing|SF6"] = "W/m**2"
    unit["Effective Radiative Forcing|CFC11"] = "W/m**2"
    unit["Effective Radiative Forcing|CFC12"] = "W/m**2"
    unit["Effective Radiative Forcing|CFC113"] = "W/m**2"
    unit["Effective Radiative Forcing|CFC114"] = "W/m**2"
    unit["Effective Radiative Forcing|CFC115"] = "W/m**2"
    unit["Effective Radiative Forcing|CCl4"] = "W/m**2"
    unit["Effective Radiative Forcing|CH3CCl3"] = "W/m**2"
    unit["Effective Radiative Forcing|HCFC22"] = "W/m**2"
    unit["Effective Radiative Forcing|HCFC141b"] = "W/m**2"
    unit["Effective Radiative Forcing|HCFC142b"] = "W/m**2"
    unit["Effective Radiative Forcing|Halon1211"] = "W/m**2"
    unit["Effective Radiative Forcing|Halon1202"] = "W/m**2"
    unit["Effective Radiative Forcing|Halon1301"] = "W/m**2"
    unit["Effective Radiative Forcing|Halon2402"] = "W/m**2"
    unit["Effective Radiative Forcing|CH3Br"] = "W/m**2"
    unit["Effective Radiative Forcing|CH3Cl"] = "W/m**2"
    unit["Effective Radiative Forcing|Tropospheric Ozone"] = "W/m**2"
    unit["Effective Radiative Forcing|Stratospheric Ozone"] = "W/m**2"
    unit["Effective Radiative Forcing|CH4 Oxidation Stratospheric H2O"] = "W/m**2"
    unit["Effective Radiative Forcing|Contrails"] = "W/m**2"
    unit["Effective Radiative Forcing|Aerosols|Direct Effect|SOx"] = "W/m**2"
    unit[
        "Effective Radiative Forcing|Aerosols|Direct Effect|Secondary Organic Aerosol"
    ] = "W/m**2"
    unit["Effective Radiative Forcing|Aerosols|Direct Effect|Nitrate"] = "W/m**2"
    unit["Effective Radiative Forcing|Aerosols|Direct Effect|BC"] = "W/m**2"
    unit["Effective Radiative Forcing|Aerosols|Direct Effect|OC"] = "W/m**2"
    unit["Effective Radiative Forcing|Aerosols|Indirect Effect"] = "W/m**2"
    unit["Effective Radiative Forcing|Black Carbon on Snow"] = "W/m**2"
    unit["Effective Radiative Forcing|Land-use Change"] = "W/m**2"
    unit["Effective Radiative Forcing|Volcanic"] = "W/m**2"
    unit["Effective Radiative Forcing|Solar"] = "W/m**2"
    unit["Effective Radiative Forcing"] = "W/m**2"
    unit["Effective Radiative Forcing|Anthropogenic"] = "W/m**2"
    unit["Effective Radiative Forcing|Greenhouse Gases"] = "W/m**2"
    unit["Effective Radiative Forcing|Kyoto Gases"] = "W/m**2"
    unit["Effective Radiative Forcing|CO2, CH4 and N2O"] = "W/m**2"
    unit["Effective Radiative Forcing|F-Gases"] = "W/m**2"
    unit["Effective Radiative Forcing|Montreal Protocol Halogen Gases"] = "W/m**2"
    unit["Effective Radiative Forcing|Aerosols|Direct Effect"] = "W/m**2"
    unit["Effective Radiative Forcing|Aerosols"] = "W/m**2"
    unit["Effective Radiative Forcing|Ozone"] = "W/m**2"
    unit["Surface Air Temperature Change"] = "K"
    unit["Surface Air Ocean Blended Temperature Change"] = "K"
    unit["Airborne Fraction"] = "dimensionless"
    unit["Effective Climate Feedback"] = "W/m**2/K"
    unit["Heat Content"] = "J"
    unit["Heat Content|Ocean"] = "J"
    unit["Net Energy Imbalance"] = "W/m**2"
    unit["Heat Uptake"] = "W/m**2"
    unit["Heat Uptake|Ocean"] = "W/m**2"

    nt = len(temperature)

    out = ({}, {}, nt)
    for key in output_vars:
        if key not in data:
            LOGGER.warning("%s not available from FaIR", key)
            continue

        out[0][key] = data[key]
        out[1][key] = unit[key]

    return out



"""
Progress bar wrapper
"""

from tqdm.autonotebook import tqdm

# TODO: handle configuration in a consistent manner
_default_tqdm_params = {
    "mininterval": 5,
    "unit": "it",
    "unit_scale": True,
}


def progress(*args, **kwargs):
    """
    Progress bar

    Uses ``tqdm.autonotebook`` to automatically use a native Jupyter widget
    when executing within a Jupyer Notebook.

    Parameters
    ----------
    *args
        Passed to the tqdm

    **kwargs
        Passed to the tqdm

    Returns
    -------
    tqdm.auto_notebook.tqdm
        tqdm instance with consistent configuration
    """
    kwargs = {**_default_tqdm_params, **kwargs}
    return tqdm(*args, **kwargs)



"""
Miscellaneous testing code
"""
import json

import numpy.testing as npt

try:
    import pytest

    HAS_PYTEST = True

except ImportError:
    HAS_PYTEST = False


def _check_output(  # pylint: disable=too-many-locals,too-many-branches
    res, expected_output_file, rtol, update
):
    if not HAS_PYTEST:
        raise ImportError("pytest not installed, run `pip install pytest`")

    with open(expected_output_file, "r", encoding="ascii") as filehandle:
        expected_output = json.load(filehandle)

    if update:
        updated_output = {}

    for climate_model, checks in expected_output.items():
        res_cm = res.filter(climate_model=climate_model)

        if update:
            updated_output[climate_model] = []

        for filter_kwargs, expected_val in checks:
            err_msg = f"{filter_kwargs}"

            if update:
                filter_kwargs_in = {**filter_kwargs}

            check_units = filter_kwargs.pop("unit", None)
            quantile = filter_kwargs.pop("quantile")
            res_to_check = res_cm.filter(**filter_kwargs)

            if check_units is not None:
                res_to_check = res_to_check.convert_unit(check_units)

            if res_to_check.empty:
                raise AssertionError

            res_val = float(
                res_to_check.process_over(
                    ("climate_model", "run_id"), "quantile", q=quantile
                ).values
            )

            try:
                npt.assert_allclose(
                    res_val, expected_val, rtol=rtol, err_msg=err_msg,
                )
                if update:
                    new_val = expected_val
            except AssertionError:
                if update:
                    new_val = res_val
                else:
                    raise

            if update:
                updated_output[climate_model].append((filter_kwargs_in, new_val))

    if update:
        with open(expected_output_file, "w", encoding="ascii") as file_handle:
            json.dump(updated_output, file_handle, indent=4)

        pytest.skip(f"Updated {expected_output_file}")



"""
High-level run function
"""
import logging

import scmdata

from .adapters import CICEROSCM, FAIR, MAGICC7
from .progress import progress

LOGGER = logging.getLogger(__name__)


def _check_out_config(out_config, climate_models_cfgs):
    if out_config is not None:
        unknown_models = set(out_config.keys()) - set(climate_models_cfgs.keys())
        if unknown_models:
            LOGGER.warning(
                "Found model(s) in `out_config` which are not in "
                "`climate_models_cfgs`: %s",
                unknown_models,
            )

        for key, value in out_config.items():
            if not isinstance(value, tuple):
                raise TypeError(
                    f"`out_config` values must be tuples, this isn't the case for "
                    f"climate_model: '{key}'"
                )


def run(
    climate_models_cfgs,
    scenarios,
    output_variables=("Surface Temperature",),
    out_config=None,
):  # pylint: disable=W9006
    """
    Run a number of climate models over a number of scenarios

    Parameters
    ----------
    climate_models_cfgs : dict[str: list]
        Dictionary where each key is a model and each value is the configs
        with which to run the model. The configs are passed to the model
        adapter.

    scenarios : :obj:`pyam.IamDataFrame`
        Scenarios to run

    output_variables : list[str]
        Variables to include in the output

    out_config : dict[str: tuple of str]
        Dictionary where each key is a model and each value is a tuple of
        configuration values to include in the output's metadata.

    Returns
    -------
    :obj:`scmdata.ScmRun`
        Model output

    Raises
    ------
    KeyError
        ``out_config`` has keys which are not in ``climate_models_cfgs``

    TypeError
        A value in ``out_config`` is not a :obj:`tuple`
    """
    _check_out_config(out_config, climate_models_cfgs)

    res = []
    for climate_model, cfgs in progress(
        climate_models_cfgs.items(), desc="Climate models"
    ):
        if climate_model == "MAGICC7":
            runner = MAGICC7()
        elif climate_model.upper() == "FAIR":  # allow various capitalisations
            runner = FAIR()
        elif climate_model.upper() == "CICEROSCM":  # allow various capitalisations
            runner = CICEROSCM()
        else:
            raise NotImplementedError(f"No adapter available for {climate_model}")

        if out_config is not None and climate_model in out_config:
            output_config_cm = out_config[climate_model]
            LOGGER.debug(
                "Using output config: %s for %s", output_config_cm, climate_model
            )
        else:
            LOGGER.debug("No output config for %s", climate_model)
            output_config_cm = None

        model_res = runner.run(
            scenarios,
            cfgs,
            output_variables=output_variables,
            output_config=output_config_cm,
        )
        res.append(model_res)

    for i, model_res in enumerate(res):
        if i < 1:
            key_meta = set(model_res.meta.columns.tolist())

        model_meta = set(model_res.meta.columns.tolist())
        climate_model = model_res.get_unique_meta("climate_model")
        if model_meta != key_meta:  # noqa
            raise AssertionError(
                f"{climate_model} meta: {model_meta}, expected meta: {key_meta}"
            )

    if len(res) == 1:
        LOGGER.info("Only one model run, returning its results")
        scmdf = res[0]
    else:
        LOGGER.info("Appending model results")
        scmdf = scmdata.run_append(res)

    return scmdf



"""
CICEROSCM adapter
"""
import logging
import os.path
from subprocess import check_output  # nosec

from ..base import _Adapter
from ._run_ciceroscm_parallel import run_ciceroscm_parallel

LOGGER = logging.getLogger(__name__)


class CICEROSCM(_Adapter):  # pylint: disable=too-few-public-methods
    """
    Adapter for CICEROSCM
    """

    def __init__(self):  # pylint: disable=useless-super-delegation
        """
        Initialise the CICEROSCM adapter

        """
        super().__init__()

    def _init_model(self):  # pylint: disable=arguments-differ
        pass

    def _run(self, scenarios, cfgs, output_variables, output_config):
        """
        Run the model.

        This method is the internal implementation of the :meth:`run` interface

        cfgs is a list of indices to run
        """
        if output_config is not None:
            raise NotImplementedError("`output_config` not implemented for CICERO-SCM")

        runs = run_ciceroscm_parallel(scenarios, cfgs, output_variables)
        return runs

    @classmethod
    def get_version(cls):
        """
        Get the CICEROSCM version being used by this adapter

        Returns
        -------
        str
            The CICEROSCM version id

        Raises
        ------
        OSError
            The CICERO-SCM binary cannot be run on the operating system
        """
        exec_call = os.path.join(
            os.path.dirname(__file__), "utils_templates", "run_dir", "scm_vCH4fb"
        )
        try:
            check_output(exec_call)
        except OSError as orig_exc:
            raise OSError(
                "CICERO-SCM is not available on your operating system"
            ) from orig_exc

        return "v2019vCH4"



"""
Module to keep track of MAGICC instances on disk
"""
import logging
import multiprocessing
import shutil
import tempfile

from ...settings import config
from ._compat import pymagicc

LOGGER = logging.getLogger(__name__)


class _MagiccInstances:
    def __init__(self, existing_instances):
        """
        Initialise a MAGICC instances handler

        Parameters
        ----------
        existing_instances : :obj:`multiprocessing.managers.DictProxy`
            Dictionary which can store new/existing instances across parallel processes
        """
        self.instances = existing_instances

    def __enter__(self):
        return self

    def cleanup(self):
        """
        Remove all MAGICC instances
        """
        # have to use list as can't modify dict whilst iterating
        insts = list(self.instances.keys())

        for magicc_inst in insts:
            LOGGER.info("removing %s", self.instances[magicc_inst].root_dir)
            shutil.rmtree(self.instances[magicc_inst].root_dir)
            self.instances.pop(magicc_inst)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()

    def __iter__(self):
        return self.instances.__iter__()

    def __getitem__(self, item):
        return self.instances[item]

    @staticmethod
    def _get_key():
        return multiprocessing.current_process().name

    @staticmethod
    def _generate_magicc_root(root_dir):
        return tempfile.mkdtemp(prefix="pymagicc-", dir=root_dir)

    def get(self, root_dir=None, init_callback=None, init_callback_kwargs=None):
        """
        Get a MAGICC object which is ready to run (always uses ``strict=False``)

        This caches the magicc instance used to minimise overhead from copying files.
        Each process gets a unique copy of MAGICC to ensure that each process has
        exclusive access to the magicc instance.

        Parameters
        ----------
        root_dir : str
            Root directory in which to create MAGICC instances.

        init_callback : func
            Function to call when making the MAGICC instance. Must have the
            same function signature as ``default_magicc_setup``.

        init_callback_kwargs : dict
            Keyword arguments to pass to :func:`init_callback`

        Returns
        -------
        pymagicc.MAGICC7
            MAGICC7 object with a valid configuration
        """
        magicc_version = 7  # hard-code for now
        key = (magicc_version, self._get_key())
        try:
            return self.instances[key]
        except KeyError:
            kwargs_init = {}
            if root_dir:
                kwargs_init["root_dir"] = self._generate_magicc_root(root_dir)

            # ensure pymagicc will behave itself
            pymagicc.config.config["EXECUTABLE_7"] = config["MAGICC_EXECUTABLE_7"]
            magicc = pymagicc.MAGICC7(strict=False, **kwargs_init)
            LOGGER.info("Creating new magicc instance: %s - %s", key, magicc.root_dir)
            magicc.create_copy()

            self.instances[key] = magicc
            if init_callback_kwargs is None:
                init_callback_kwargs = {}

            if init_callback is not None:
                init_callback(magicc, **init_callback_kwargs)

            return magicc
